#!/usr/bin/env python
# -*- coding: utf-8 -*-

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides the set of base classes from which the Qt taurus widgets
should inherit to be considered valid taurus widgets."""

__all__ = ["TaurusBaseComponent", "TaurusBaseWidget",
           "TaurusBaseWritableWidget"]

__docformat__ = 'restructuredtext'

import sys
import threading

from taurus.external.qt import Qt
from taurus.external.enum import Enum

import taurus
from taurus.core.util import eventfilters
from taurus.core.util.timer import Timer
from taurus.core.taurusbasetypes import TaurusElementType, TaurusEventType
from taurus.core.taurusattribute import TaurusAttribute
from taurus.core.taurusdevice import TaurusDevice
from taurus.core.taurusconfiguration import (TaurusConfiguration,
                                             TaurusConfigurationProxy)
from taurus.core.tauruslistener import TaurusListener, TaurusExceptionListener
from taurus.core.taurusoperation import WriteAttrOperation
from taurus.core.util.eventfilters import filterEvent
from taurus.core.util.log import deprecation_decorator
from taurus.qt.qtcore.util.signal import baseSignal
from taurus.qt.qtcore.configuration import BaseConfigurableClass
from taurus.qt.qtcore.mimetypes import TAURUS_ATTR_MIME_TYPE, TAURUS_DEV_MIME_TYPE, TAURUS_MODEL_MIME_TYPE
from taurus.qt.qtgui.util import ActionFactory

DefaultNoneValue = "-----"


class TaurusBaseComponent(TaurusListener, BaseConfigurableClass):
    """A generic Taurus component.

       .. note::
           Any class which inherits from TaurusBaseComponent is expected to also
           inherit from QObject (or from a QObject derived class).
       .. note::
           :meth:`getSignaller` is now unused and deprecated. This is because
           `taurusEvent` is implemented using :func:`baseSignal`, that doesn't
           require the class to inherit from QObject.

    """
    _modifiableByUser = False
    _showQuality = True
    _eventBufferPeriod = 0

    taurusEvent = baseSignal('taurusEvent', object, object, object)

    def __init__(self, name, parent=None, designMode=False):
        """Initialization of TaurusBaseComponent"""
        self.modelObj = None
        self.modelName = ''
        self.modelFragmentName = None
        self.noneValue = DefaultNoneValue
        self._designMode = designMode
        self.call__init__(TaurusListener, name, parent)

        BaseConfigurableClass.__init__(self)

        self.taurusMenu = None
        self.taurusMenuData = ''

        # attributes storing property values
        self._localModelName = ''
        self._useParentModel = False
        self._showText = True
        self._attached = False
        self._dangerMessage = ""
        self._isDangerous = False
        self._forceDangerousOperations = False
        self._eventFilters = []
        self._preFilters = []
        self._isPaused = False
        self._operations = []
        self._modelInConfig = False
        self._autoProtectOperation = True

        self._bufferedEvents = {}
        self._bufferedEventsTimer = None
        self.setEventBufferPeriod(self._eventBufferPeriod)

        if parent is not None and hasattr(parent, "_exception_listener"):
            self._exception_listener = parent._exception_listener
        else:
            self._exception_listener = set([TaurusExceptionListener()])

        # register configurable properties
        self.registerConfigProperty(
            self.isModifiableByUser, self.setModifiableByUser, "modifiableByUser")
        self.registerConfigProperty(
            self.getModelInConfig, self.setModelInConfig, "ModelInConfig")
        self.resetModelInConfig()

    @deprecation_decorator(rel='4.0')
    def getSignaller(self):
        return self

    def deleteLater(self):
        '''Reimplements the Qt.QObject deleteLater method to ensure that the
        this object stops listening its model.'''
        self.setUseParentModel(False)
        self.resetModel()
        Qt.QObject.deleteLater(self)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Utility methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getTaurusManager(self):
        """Returns the taurus manager singleton. This is just a helper method.
        It is the equivalent of doing::

            import taurus
            manager = taurus.Manager()

        :return: (taurus.core.taurusmanager.TaurusManager) the TaurusManager
        """
        return taurus.Manager()

    def getTaurusFactory(self, scheme=''):
        """Returns the taurus factory singleton for the given scheme.
        This is just a helper method. It is the equivalent of doing::

            import taurus
            factory = taurus.Factory(scheme)

        :param scheme: (str or None) the scheme. If scheme is an empty string,
                       or is not passed, the scheme will be obtained from the
                       model name. For backwards compatibility (but deprecated),
                       passing None is equivalent to 'tango'.

        :return: (taurus.core.taurusfactory.TaurusFactory) the TaurusFactory
        """
        if scheme == '':
            scheme = taurus.getSchemeFromName(self.getModelName() or '')
        if scheme is None:
            scheme = 'tango'

        return taurus.Factory(scheme)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Popup menu behavior
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def contextMenuEvent(self, event):
        """Handle the popup menu event

        :param event: the popup menu event
        """
        if self.taurusMenu is not None:
            self.taurusMenu.exec_(event.globalPos())
        else:
            event.ignore()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Mandatory methods to be implemented in subclass implementation
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~

    def updateStyle(self):
        """Method called when the component detects an event that triggers a
        change in the style.
        Default implementation doesn't do anything. Overwrite when necessary
        """
        pass

    def getParentTaurusComponent(self):
        """ Returns a parent Taurus component or None if no parent
        :class:`taurus.qt.qtgui.base.TaurusBaseComponent` is found.

        :raises: RuntimeError
        """
        raise RuntimeError(
            "Not allowed to call TaurusBaseComponent::getParentTaurusComponent()")

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Event handling chain
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setEventBufferPeriod(self, period):
        '''Set the period at wich :meth:`fireBufferedEvents` will be called.
        If period is 0, the event buffering is disabled (i.e., events are fired
        as soon as they are received)

        :param period: (float) period in seconds for the automatic event firing.
                    period=0 will disable the event buffering.
        '''
        self._eventBufferPeriod = period
        if period == 0:
            if self._bufferedEventsTimer is not None:
                self._bufferedEventsTimer.stop()
                self._bufferedEventsTimer = None
                self.fireBufferedEvents()  # flush the buffer
        else:
            self._eventsBufferLock = threading.RLock()
            self._bufferedEventsTimer = Timer(period, self.fireBufferedEvents,
                                              self)
            self._bufferedEventsTimer.start()

    def getEventBufferPeriod(self):
        '''Returns the event buffer period

        :return: (float) period (in s). 0 means event buffering is disabled.
        '''
        return self._eventBufferPeriod

    def eventReceived(self, evt_src, evt_type, evt_value):
        """The basic implementation of the event handling chain is as
        follows:

            - eventReceived just calls :meth:`fireEvent` which emits a "taurusEvent"
              PyQt signal that is connected (by :meth:`preAttach`) to the
              :meth:`filterEvent` method.
            - After filtering, :meth:`handleEvent` is invoked with the resulting
              filtered event

        .. note::
            in the earlier steps of the chain (i.e., in :meth:`eventReceived`/
            :meth:`fireEvent`), the code is executed in a Python thread, while
            from eventFilter ahead, the code is executed in a Qt thread.
            When writing widgets, one should normally work on the Qt thread
            (i.e. reimplementing :meth:`handleEvent`)

        :param evt_src: (object) object that triggered the event
        :param evt_type: (taurus.core.taurusbasetypes.TaurusEventType) type of event
        :param evt_value: (object) event value
        """
        evt = filterEvent(evt_src, evt_type, evt_value,
                          filters=self._preFilters)
        if evt is not None:
            self.fireEvent(*evt)

    def fireEvent(self, evt_src=None, evt_type=None, evt_value=None):
        """Emits a "taurusEvent" signal.
        It is unlikely that you need to reimplement this method in subclasses.
        Consider reimplementing :meth:`eventReceived` or :meth:`handleEvent`
        instead depending on whether you need to execute code in the python
        or Qt threads, respectively

        :param evt_src: (object or None) object that triggered the event
        :param evt_type: (taurus.core.taurusbasetypes.TaurusEventType or None)
                         type of event
        :param evt_value: (object or None) event value
        """
        if self._eventBufferPeriod:
            # If we have an active event buffer delay, store the event...
            with self._eventsBufferLock:
                self._bufferedEvents[(evt_src, evt_type)] = (evt_src, evt_type,
                                                             evt_value)
        else:
            # if we are not buffering, directly emit the signal
            try:
                self.taurusEvent.emit(evt_src, evt_type, evt_value)
            except:
                pass  # self.error('%s.fireEvent(...) failed!'%type(self))

    def fireBufferedEvents(self):
        '''Fire all events currently buffered (and flush the buffer)

        Note: this method is normally called from an event buffer timer thread
              but it can also be called any time the buffer needs to be flushed
        '''
        with self._eventsBufferLock:
            for evt in self._bufferedEvents.values():
                self.taurusEvent.emit(*evt)
            self._bufferedEvents = {}

    def filterEvent(self, evt_src=-1, evt_type=-1, evt_value=-1):
        """The event is processed by each and all filters in strict order
        unless one of them returns None (in which case the event is discarded)

        :param evt_src: (object) object that triggered the event
        :param evt_type: (taurus.core.taurusbasetypes.TaurusEventType) type of event
        :param evt_value: (object) event value
        """
        evt = evt_src, evt_type, evt_value

        if evt == (-1, -1, -1):
            # @todo In an ideal world the signature of this method should be
            # (evt_src, evt_type, evt_value). However there's a bug in PyQt:
            # If a signal is disconnected between the moment it is emitted and
            # the moment the slot is called, then the slot is called without
            # parameters (!?). We added this default values to detect if
            # this is the case without printing an error message each time.
            # If this gets fixed, we should remove this line.
            return

        evt = filterEvent(*evt, filters=self._eventFilters)
        if evt is not None:
            self.handleEvent(*evt)

    def handleEvent(self, evt_src, evt_type, evt_value):
        """Event handling. Default implementation does nothing.
        Reimplement as necessary

        :param evt_src: (object or None) object that triggered the event
        :param evt_type: (taurus.core.taurusbasetypes.TaurusEventType or None) type of event
        :param evt_value: (object or None) event value
        """
        pass

    def setEventFilters(self, filters=None, preqt=False):
        """sets the taurus event filters list.
        The filters are run in order, using each output to feed the next filter.
        A filter must be a function that accepts 3 arguments ``(evt_src, evt_type, evt_value)``
        If the event is to be ignored, the filter must return None.
        If the event is  not to be ignored, filter must return a
        ``(evt_src, evt_type, evt_value)`` tuple which may (or not) differ from the input.

        For a library of common filters, see taurus/core/util/eventfilters.py

        :param filters: (sequence) a sequence of filters
        :param preqt: (bool) If true, set the pre-filters (that are applied in
                      eventReceived, at the python thread),
                      otherwise, set the filters to be applied at the main
                      Qt thread (default)

        *Note*: If you are setting a filter that applies a transformation on
        the parameters, you may want to generate a fake event to force the last
        value to be filtered as well. This can be done as in this example::

            TaurusBaseComponent.fireEvent( TaurusBaseComponent.getModelObj(),
                                        taurus.core.taurusbasetypes.TaurusEventType.Periodic,
                                        TaurusBaseComponent.getModelObj().getValueObj())

        See also: insertEventFilter
        """
        if filters is None:
            filters = []
        if preqt:
            self._preFilters = list(filters)
        else:
            self._eventFilters = list(filters)

    def getEventFilters(self, preqt=False):
        """Returns the list of event filters for this widget

        :param preqt: (bool) If true, return the pre-filters (that are applied
                      in eventReceived, at the python thread),
                      otherwise, return the filters to be applied at the main
                      Qt thread (default)

        :return: (sequence<callable>) the event filters
        """
        return (self._preFilters if preqt else self._eventFilters)

    def insertEventFilter(self, filter, index=-1, preqt=False):
        """insert a filter in a given position

        :param filter: (callable(evt_src, evt_type, evt_value)) a filter
        :param index: (int) index to place the filter (default = -1 meaning place at the end)
        :param preqt: (bool) If true, set the pre-filters (that are applied in
                      eventReceived, at the python thread),
                      otherwise, set the filters to be applied at the main
                      Qt thread (default)


        See also: setEventFilters
        """
        if preqt:
            self._preFilters.insert(index, filter)
        else:
            self._eventFilters.insert(index, filter)

    def setPaused(self, paused=True):
        """Toggles the pause mode.

        :param paused: (bool) whether or not to pause (default = True)
        """
        if paused == self._isPaused:
            return  # nothing to do
        if paused:  # pausing
            self.insertEventFilter(eventfilters.IGNORE_ALL, 0)
            self.debug('paused')
        else:  # unpausing
            try:
                self._eventFilters.remove(eventfilters.IGNORE_ALL)
                self.debug('Unpaused')
            except ValueError:
                self.warning('Unpause failed')
        self._isPaused = paused

    def isPaused(self):
        """Return the current pause state

        :return: (bool) wheater or not the widget is paused
        """
        return self._isPaused

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Model class methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getModelClass(self):
        """Return the class object for the widget.
        Default behavior is to do a 'best effort' to determine which model
        type corresponds to the current model name.
        Overwrite as necessary.

        :return: (class TaurusModel or None) The class object corresponding to the type
                 of Taurus model this widget handles or None if no valid class is found.
        """
        return self.findModelClass()

    def findModelClass(self):
        """Do a "best effort" to determine which model type corresponds to the
        given model name.

        :return: (class TaurusModel or None) The class object corresponding to the type
                 of Taurus model this widget handles or None if no valid class is found.
        """
        if self.getUseParentModel():
            return self._findRelativeModelClass(self.getModel())
        else:
            return self._findAbsoluteModelClass(self.getModel())

    def _findAbsoluteModelClass(self, absolute_name):
        return taurus.Manager().findObjectClass(absolute_name)

    def _findRelativeModelClass(self, relative_name):
        parent_widget = self.getParentTaurusComponent()
        if parent_widget is None:
            return None

        parent_obj = parent_widget.getModelObj()
        if parent_obj is None:
            return None

        if relative_name is None or len(relative_name) == 0:
            return parent_widget.getModelClass()

        obj = parent_obj.getChildObj(relative_name)
        if obj is None:
            return None
        if isinstance(obj, TaurusConfigurationProxy):
            return obj.getRealConfigClass()
        else:
            return obj.__class__

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Model related methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setModelName(self, modelName, parent=None):
        """This method will detach from the previous taurus model (if any), it
        will set the new model to the given modelName and it will attach
        this component to the new taurus model.

        :param modelName: (str) the new taurus model name (according to the taurus convention)
        :param parent: (TaurusBaseComponent) the parent or None (default) if this
                       component does not have a parent Taurus component
        """
        modelName = str(modelName)
        if parent:
            modelClass = self.getModelClass()
            if modelClass is not None:
                parent_model = self.getParentModelObj()
                modelName = modelClass.buildModelName(parent_model, modelName)
        self._detach()
        self.modelName = modelName
        # update modelFragmentName
        try:
            scheme = self.getTaurusManager().getScheme(modelName)
            factory = taurus.Factory(scheme)
            v = factory.getAttributeNameValidator()
            self.modelFragmentName = v.getUriGroups(self.modelName)['fragment']
        except (AttributeError, TypeError):
            self.modelFragmentName = None
        self._attach()

    def getModelName(self):
        """Returns the current model name.

        :return: (str) the model name
        """
        return self.modelName

    def getFullModelName(self):
        """Returns the full name of the current model object.

        :return: (str) the model name
        """
        obj = self.getModelObj()
        if obj is None:
            return None
        return obj.getFullName()

    def getParentModelName(self):
        """Returns the parent model name or an empty string if the component
        has no parent

        :return: (str) the parent model name
        """
        try:
            p = self.getParentTaurusComponent()
            if p is None:
                return ''
        except:
            return ''
        return p.getModelName()

    def getParentModelObj(self):
        """Returns the parent model object or None if the component has no
        parent or if the parent's model is None

        :return: (taurus.core.taurusmodel.TaurusModel or None) the parent taurus model object
        """
        try:
            p = self.getParentTaurusComponent()
            if p is None:
                return None
        except:
            return None
        return p.getModelObj()

    def getModelObj(self):
        """Returns the taurus model obj associated with this component or None if
        no taurus model is associated.

        :return: (taurus.core.taurusmodel.TaurusModel or None) the taurus model object
        """
        return self.modelObj

    def getModelType(self):
        """Returns the taurus model type associated with this component or
        taurus.core.taurusbasetypes.TaurusElementType.Unknown if no taurus model is associated.

        :return: (taurus.core.taurusbasetypes.TaurusElementType) the taurus model type
        """
        model_obj = self.getModelObj()
        if model_obj is None:
            return TaurusElementType.Unknown
        return model_obj.getTaurusElementType()

    def getModelValueObj(self, cache=True):
        """Returns the tango obj value associated with this component or None
        if no taurus model is associated.

        :param cache: (bool) if set to True (default) use the cache value. If set to
                      False will force a connection to the server.

        :return: (TaurusAttrValue) the tango value object.
        """
        if self.modelObj is None:
            return None
        return self.modelObj.getValueObj(cache=cache)

    def getModelFragmentObj(self, fragmentName=None):
        """Returns a fragment object of the model. A fragment of a model is a
        python attribute of the model object.

        Fragment names including dots will be used to recursively get fragments
        of fragments.

        For a simple fragmentName (no dots), this is roughly equivalent to
        getattr(self.getModelObj(), fragmentName)

        If the model does not have that fragment, :class:`AttributeError` is
        raised (other exceptions may be raised when accessing the fragment as
        well)

        :param fragmentName: (str or None) the returned value will correspond to
                         the given fragmentName. If None passed,
                         self.modelFragmentName will be used, and if None is
                         set, the defaultFragmentName of the model will be used
                         instead.

        :return: (obj) the member of the modelObj referred by the fragment.
        """
        if fragmentName is None:  # no fragmentName passed in kwargs
            fragmentName = self.modelFragmentName
        return self.modelObj.getFragmentObj(fragmentName)

    def getFormatedToolTip(self, cache=True):
        """Returns a string with contents to be displayed in a tooltip.

        :param cache: (bool) if set to True (default) use the cache value. If set to
                      False will force a connection to the server.

        :return: (str) a tooltip
        """
        if self.modelObj is None:
            return self.getNoneValue()
        obj = self.modelObj.getDisplayDescrObj()
        return self.toolTipObjToStr(obj)

    def toolTipObjToStr(self, toolTipObj):
        """Converts a python dict to a tooltip string.

        :param toolTipObj: (dict) a python object

        :return: (str) a tooltip
        """
        if toolTipObj is None:
            return self.getNoneValue()
        ret = '<TABLE width="500" border="0" cellpadding="1" cellspacing="0">'

        for id, value in toolTipObj:
            ret += '<TR><TD WIDTH="80" ALIGN="RIGHT" VALIGN="MIDDLE"><B>%s:</B></TD><TD>%s</TD></TR>' % (
                id.capitalize(), value)
        ret += '</TABLE>'
        return ret

    def displayValue(self, v):
        """Returns a string representation of the given value

        :param v: (object) the value to be translated to string

        :return: (str) a string representing the given value
        """
        if isinstance(v, Enum):
            return v.name
        return str(v)

    def getDisplayValue(self, cache=True, fragmentName=None):
        """Returns a string representation of the model value associated with
        this component.

        :param cache: (bool) (ignored, just for bck-compat).
        :param fragmentName: (str or None) the returned value will correspond to
                         the given fragmentName. If None passed,
                         self.modelFragmentName will be used, and if None is
                         set, the defaultFragmentName of the model will be used
                         instead.

        :return: (str) a string representation of the model value.
        """
        if self.modelObj is None:
            return self.getNoneValue()
        try:
            v = self.getModelFragmentObj(fragmentName=fragmentName)
        except:
            return self.getNoneValue()
        return self.displayValue(v)

    def setNoneValue(self, v):
        """Sets the new string representation when no model or no model value exists.

        :param v: (str) the string representation for an invalid value
        """
        self.noneValue = v

    def getNoneValue(self):
        """Returns the current string representation when no valid model or model value exists.

        :return: (str) a string representation for an invalid value
        """
        return self.noneValue

    def isChangeable(self):
        """Tells if this component value can be changed by the user. Default implementation
        will return True if and only if:

            - this component is attached to a valid taurus model and
            - the taurus model is writable and
            - this component is not read-only

        :return: (bool) True if this component value can be changed by the user or False otherwise
        """
        res = False
        if not self.modelObj is None:
            res = self.modelObj.isWritable()
        res = res and not self.isReadOnly()
        return res

    def isReadOnly(self):
        """Determines if this component is read-only or not in the sense that the
        user can interact with it. Default implementation returns True.

        Override when necessary.

        :return: (bool) whether or not this component is read-only
        """
        return True

    def isAttached(self):
        """Determines if this component is attached to the taurus model.

        :return: (bool) True if the component is attached or False otherwise.
        """
        return self._attached

    def preAttach(self):
        """Called inside self.attach() before actual attach is performed.
        Default implementation just emits a signal.

        Override when necessary.
        """
        try:
            self.taurusEvent.connect(self.filterEvent)
        except:
            # self.error("In %s.preAttach() ... failed!" % str(type(self)))
            pass

    def postAttach(self):
        """Called inside self.attach() after actual attach is performed.
        Default implementation does not do anything.

        Override when necessary.
        """
        pass

    def preDetach(self):
        """Called inside self.detach() before actual deattach is performed.
        Default implementation just disconnects a signal.

        Override when necessary.
        """
        try:
            self.taurusEvent.disconnect(self.filterEvent)
        except:
            # self.error("In %s.preDetach() ... failed!" % str(type(self)))
            pass

    def postDetach(self):
        """Called inside self.detach() after actual deattach is performed.
        Default implementation does not do anything.

        Override when necessary.
        """
        pass

    def _attach(self):
        """Attaches the component to the taurus model.
        In general it should not be necessary to overwrite this method in a
        subclass.

        :return: (bool) True if success in attachment or False otherwise.
        """
        if self.isAttached():
            return self._attached

        self.preAttach()

        cls = self.getModelClass()

        if cls is None:
            self._attached = False
            #self.trace("Failed to attach: Model class not found")
        elif self.modelName == '':
            self._attached = False
            self.modelObj = None
        else:
            try:
                self.modelObj = taurus.Manager().getObject(cls, self.modelName)
                if self.modelObj is not None:
                    self.modelObj.addListener(self)
                    self._attached = True
                    self.changeLogName(self.log_name + "." + self.modelName)
            except Exception:
                self.modelObj = None
                self._attached = False
                self.debug(
                    "Exception occured while trying to attach '%s'" % self.modelName)
                self.traceback()

        self.postAttach()
        return self._attached

    def _detach(self):
        """Detaches the component from the taurus model"""
        self.preDetach()

        if self.isAttached():
            m = self.getModelObj()
            if not m is None:
                m.removeListener(self)

            pos = self.log_name.find('.')
            if pos >= 0:
                new_log_name = self.log_name[:self.log_name.rfind('.')]
                self.changeLogName(new_log_name)
            self.modelObj = None
            self._attached = False
            self.fireEvent(m, TaurusEventType.Change, None)

        self.postDetach()

    def setModelInConfig(self, yesno):
        '''
        Sets whether the model-related properties should be stored for this
        widget when creating the config dict with :meth:`createConfig` (and
        restored when calling :meth:`applyConfig`).
        By default this is not enabled.
        The following properties are affected by this:
        - "model"

        :param yesno: (bool) If True, the model-related properties will be
                      registered as config properties. If False, they will be
                      unregistered.

        .. seealso:: :meth:`registerConfigProperty`, :meth:`createConfig`,
                     :meth:`applyConfig`

        '''
        if yesno == self._modelInConfig:
            return
        if yesno:
            self.registerConfigProperty(self.getModel, self.setModel, "model")
        else:
            self.unregisterConfigurableItem("model", raiseOnError=False)
        self._modelInConfig = yesno

    def getModelInConfig(self):
        return self._modelInConfig

    def resetModelInConfig(self):
        return self.setModelInConfig(False)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Pending operations related methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def applyPendingOperations(self, ops=None):
        """Applies operations without caring about danger messages.
        Use :meth:`TaurusBaseWidget.safeApplyOperation` if you want to warn the
        user before applying

        :param ops: (sequence<taurus.core.taurusoperation.TaurusOperation> or None) list of operations to apply.
                    If None is given (default) the component fetches the pending operations
        """
        self.debug("Apply changes")
        if ops is None:
            ops = self.getPendingOperations()

        if self.isAutoProtectOperation():
            import taurus.qt.qtgui.dialog

            @taurus.qt.qtgui.dialog.protectTaurusMessageBox
            def go():
                self.getTaurusManager().applyPendingOperations(ops)
            go()
        else:
            self.getTaurusManager().applyPendingOperations(ops)

    def hasPendingOperations(self):
        """Returns if the component has pending operations

        :return: (bool) True if there are pending operations or False otherwise
        """
        return len(self.getPendingOperations()) > 0

    def getPendingOperations(self):
        """Returns the sequence of pending operations

        :return:  (sequence<taurus.core.taurusoperation.TaurusOperation>) a list of pending operations
        """
        return self._operations

    def resetPendingOperations(self):
        """Clears the list of pending operations"""
        self._operations = []

    def setDangerMessage(self, dangerMessage=""):
        """Sets the danger message when applying an operation. If dangerMessage is None,
        the apply operation is considered safe

        :param dangerMessage: (str or None) the danger message. If None is given (default)
                              the apply operation is considered safe
        """
        self._dangerMessage = dangerMessage
        self._isDangerous = len(dangerMessage) > 0

    def getDangerMessage(self):
        """Returns the current apply danger message or None if the apply operation is safe

        :return: (str or None) the apply danger message
        """
        return self._dangerMessage

    def resetDangerMessage(self):
        """Clears the danger message. After this method is executed the apply operation
        for this component will be considered safe."""
        self.setDangerMessage(None)

    def isDangerous(self):
        """Returns if the apply operation for this component is dangerous

        :return: (bool) wheter or not the apply operation for this component is dangerous
        """
        return self._isDangerous

    def setForceDangerousOperations(self, yesno):
        """Forces/clears the dangerous operations

        :param yesno: (bool) force or not the dangerous operations"""
        self._forceDangerousOperations = yesno

    def getForceDangerousOperations(self):
        """Returns if apply dangerous operations is forced

        :return: (bool) wheter or not apply dangerous operations is forced
        """
        return self._forceDangerousOperations

    def resetForceDangerousOperations(self):
        """Clears forcing apply dangerous operations"""
        self.setForceDangerousOperations(False)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Standard Qt properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getModel(self):
        """Returns the model name for this component.

        :return: (str) the model name.
        """
        return self._localModelName

    def setModel(self, model):
        """Sets/unsets the model name for this component

        :param model: (str) the new model name"""
        self.setModelCheck(model)
        self.updateStyle()

    def setModelCheck(self, model, check=True):
        """Sets the component taurus model. Setting the check argument to True
        (default) will check if the current model is equal to the given argument.
        If so then nothing is done. This should be the standard way to call this
        method since it will avoid recursion.

        :param model: (str) the new model name
        :param check: (bool) whether or not to check against the actual model name"""
        if model is None:
            model = ''
        model = str(model)
        if check == True and model == self._localModelName:
            return

        self._localModelName = model

        parent_widget = None
        try:
            # if this widget has a buddy, check to see if it is a valid
            # TaurusWidget
            buddy_func = getattr(self, 'buddy')
            buddy_widget = buddy_func()
            if buddy_widget and isinstance(buddy_widget, TaurusBaseComponent):
                parent_widget = buddy_widget
            elif self.getUseParentModel():
                parent_widget = self.getParentTaurusComponent()
        except:
            if self.getUseParentModel():
                parent_widget = self.getParentTaurusComponent()
        self.setModelName(model, parent_widget)
        #self.fireEvent(self.getModelObj(), taurus.core.taurusbasetypes.TaurusEventType.Change, self.getModelValueObj())

    def resetModel(self):
        """Sets the model name to the empty string"""
        self.setModel('')

    def getUseParentModel(self):
        """Returns whether this component is using the parent model

        :return: (bool) True if using parent model or False otherwise
        """
        return getattr(self, '_useParentModel', False)

    @Qt.pyqtSlot(bool)
    def setUseParentModel(self, yesno):
        """Sets/unsets using the parent model

        :param yesno: (bool) whether or not to use parent model
        """
        if yesno == self._useParentModel:
            return
        self._useParentModel = yesno
        # force a recalculation of the model
        self.setModelCheck(self.getModel(), False)

    def resetUseParentModel(self):
        """Resets the usage of parent model to False"""
        self.setUseParentModel(False)
        self.updateStyle()

    @Qt.pyqtSlot(bool)
    def setShowQuality(self, showQuality):
        """Sets/unsets the show quality property

        :param showQuality: (bool) whether or not to show the quality
        """
        if showQuality == self._showQuality:
            return
        self._showQuality = showQuality
        self.updateStyle()

    def getShowQuality(self):
        """Returns if showing the quality as a background color

        :return: (bool) True if showing the quality or False otherwise
        """
        return self._showQuality

    def resetShowQuality(self):
        """Resets the show quality to self.__class__._showQuality"""
        self.setShowQuality(self.__class__._showQuality)

    @Qt.pyqtSlot(bool)
    def setShowText(self, showText):
        """Sets/unsets showing the display value of the model

        :param showText: (bool) whether or not to show the display value
        """
        if showText == self._showText:
            return
        self._showText = showText
        self.fireEvent(self.getModelObj(), TaurusEventType.Change,
                       self.getModelValueObj())
        self.updateStyle()

    def getShowText(self):
        """Returns if showing the display value

        :return: (bool) True if showing the display value or False otherwise
        """
        return self._showText

    def resetShowText(self):
        """Resets the showing of the display value to True"""
        self.setShowText(True)

    def setTaurusPopupMenu(self, menuData):
        """Sets/unsets the taurus popup menu

        :param menuData: (str) an xml representing the popup menu"""
        self.taurusMenuData = str(menuData)
        factory = ActionFactory()
        self.taurusMenu = factory.getNewMenu(self, self.taurusMenuData)

    def getTaurusPopupMenu(self):
        """Returns an xml string representing the current taurus popup menu

        :return: (str) an xml string representing the current taurus popup menu
        """
        return self.taurusMenuData

    def resetTaurusPopupMenu(self):
        """Resets the taurus popup menu to empty"""
        self.taurusMenuData = ''

    def isModifiableByUser(self):
        '''whether the user can change the contents of the widget

        :return: (bool) True if the user is allowed to modify the look&feel'''
        return self._modifiableByUser

    def setModifiableByUser(self, modifiable):
        '''
        sets whether the user is allowed to modify the look&feel

        :param modifiable: (bool)
        '''
        self._modifiableByUser = modifiable

    def resetModifiableByUser(self):
        '''Equivalent to setModifiableByUser(self.__class__._modifiableByUser)'''
        self.setModifiableByUser(self.__class__._modifiableByUser)

    def resetAutoProtectOperation(self):
        """Resets protecting operations"""
        self.setAutoProtectOperation(True)

    def isAutoProtectOperation(self):
        """Tells if this widget's operations are protected against exceptions

        :return: (bool) True if operations are protected against exceptions or
                 False otherwise"""
        return self._autoProtectOperation

    def setAutoProtectOperation(self, protect):
        """Sets/unsets this widget's operations are protected against exceptions

        :param protect: wheater or not to protect widget operations
        :type protect: bool"""
        self._autoProtectOperation = protect


class TaurusBaseWidget(TaurusBaseComponent):
    """The base class for all Qt Taurus widgets.

    .. note::
        Any class which inherits from TaurusBaseWidget is expected to also
        inherit from QWidget (or from a QWidget derived class)"""

    modelChanged = baseSignal('modelChanged', 'QString')
    valueChangedSignal = baseSignal('valueChanged')

    _dragEnabled = False

    def __init__(self, name, parent=None, designMode=False):
        self._disconnect_on_hide = False
        self._supportedMimeTypes = None
        self._autoTooltip = True
        self.call__init__(TaurusBaseComponent, name,
                          parent=parent, designMode=designMode)
        self._setText = self._findSetTextMethod()

    # It makes the GUI to hang... If this needs implementing, we should
    # reimplement it using the Qt parent class, not QWidget...
    # def destroy(self):
    #    '''Reimplements the Qt.QWidget destroy method to ensure that this object
    #    stops listening its model.'''
    #    self.setUseParentModel(False)
    #    self.resetModel()
    #    Qt.QWidget.destroy(self)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Helper methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getQtClass(self, bases=None):
        """Returns the parent Qt class for this widget

        :param bases: (sequence<class> or None) the list of class objects. If None
                      is given (default) it uses the object base classes from __bases__

        :return: (QWidget class) the QWidget class object
        """
        bases = bases or self.__class__.__bases__
        for klass in bases:
            is_taurusbasewidget = issubclass(klass, TaurusBaseWidget)
            if issubclass(klass, Qt.QWidget):
                if is_taurusbasewidget:
                    return self.getQtClass(klass.__bases__)
                return klass
            elif is_taurusbasewidget:
                return self.getQtClass(klass.__bases__)
        return None

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Qt properties from TaurusBaseComponent that need to be overwritten
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    _UseParentMsg = False

    @Qt.pyqtSlot(bool)
    def setUseParentModel(self, yesno):
        """Sets/unsets using the parent model.

        .. note:: in some cases you may need to call :meth:`recheckTaurusParent`
                  after reparenting of some of this widget's ancestors

        :param yesno: (bool) whether or not to use parent model

        .. seealso:: :meth:`recheckTaurusParent`
        """
        is_same = yesno == self._useParentModel
        if not is_same:
            self._updateUseParentModel(yesno)
            if yesno and self._designMode and not TaurusBaseWidget._UseParentMsg:
                TaurusBaseWidget._UseParentMsg = True
                Qt.QMessageBox.information(self, "UseParentModel usage note",
                                           "Using the UseParentModel feature may require you to call " +
                                           "recheckTaurusParent() manually for this widget after calling " +
                                           "setupUi in your code." +
                                           "See the documentation of TaurusBaseWidget.recheckTaurusParent()")
        TaurusBaseComponent.setUseParentModel(self, yesno)

    def _updateUseParentModel(self, yesno):
        parent_widget = self.getParentTaurusComponent()
        if parent_widget:
            if yesno:
                parent_widget.modelChanged.connect(self.parentModelChanged)
            else:
                parent_widget.modelChanged.disconnect(self.parentModelChanged)

    def recheckTaurusParent(self):
        '''
        Forces the widget to recheck its Taurus parent. Taurus Widgets will in most
        situations keep track of changes in their taurus parenting, but in some
        special cases (which unfortunately tend to occur when using Qt
        Designer) they may not update it correctly.

        If this happens, you can manually call this method.

        For more information, check the :download:`issue demo example
        </devel/examples/parentmodel_issue_demo.py>`
        '''
        self._updateUseParentModel(True)

    def setModelCheck(self, model, check=True):
        """Sets the component taurus model. Setting the check argument to True
        (default) will check if the current model is equal to the given argument.
        If so then nothing is done. This should be the standard way to call this
        method since it will avoid recursion.

        :param model: (str) the new model name
        :param check: (bool) whether or not to check against the actual model name
        """
        if model is None:
            model = ''
        model = str(model)
        send_signal = (model != self._localModelName)
        TaurusBaseComponent.setModelCheck(self, model, check)

        if send_signal:
            # emit a signal informing the child widgets that the model has
            # changed
            self.modelChanged.emit(model)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Default Qt signal handlers. Overwrite them as necessary
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def changeEvent(self, evt):
        """overwrites QWidget.changeEvent(self, evt) to handle the ParentChangeEvent
        in case this widget is using the parent model. Always calls the QWidget.changeEvent
        in order not to lose events
        """
        if self.getUseParentModel():
            evt_type = evt.type()
            if evt_type == Qt.QEvent.ParentChange:
                # disconnect from old parent
                if self._parentTaurusComponent:
                    self._parentTaurusComponent.modelChanged.disconnect(
                        self.parentModelChanged)
                self._updateUseParentModel(True)
                self.setModelCheck(self.getModel(), False)
        self.getQtClass().changeEvent(self, evt)

    def parentModelChanged(self, parentmodel_name):
        """Invoked when the Taurus parent model changes

        :param parentmodel_name: (str) the new name of the parent model
        """
        self.debug("Parent model changed to '%s'" % parentmodel_name)
        parentmodel_name = str(parentmodel_name)
        if self.getUseParentModel():
            # force an update of the interpretation of the model property
            model = self.getModel()
            self.setModelCheck(model, False)
            self.modelChanged.emit(model)
        else:
            self.debug(
                "received event from parent although not using parent model")

    def handleEvent(self, evt_src, evt_type, evt_value):
        """very basic and generalistic handling of events.

        Override when necessary.

        :param evt_src: (object or None) object that triggered the event
        :param evt_type: (taurus.core.taurusbasetypes.TaurusEventType or None) type of event
        :param evt_value: (object or None) event value
        """
        # Update the text shown by the widget
        if self._setText:
            text = ''
            if self.getShowText():
                if isinstance(evt_src, TaurusAttribute):
                    if evt_type in (TaurusEventType.Change, TaurusEventType.Periodic):
                        text = self.displayValue(evt_value.rvalue)
                    elif evt_type == TaurusEventType.Error:
                        text = self.getNoneValue()
                    elif evt_type == TaurusEventType.Config:
                        text = self.getDisplayValue()
                else:
                    text = self.getDisplayValue()
            self._setText(text)

        # update tooltip
        if self._autoTooltip:
            self.setToolTip(self.getFormatedToolTip())

        # TODO: update whatsThis

        # update appearance
        self.updateStyle()

    def setModelInConfig(self, yesno):
        '''
        extends :meth:`TaurusBaseComponent.setModelInConfig` to include also
        the "useParentModel" property

        .. seealso:: :meth:`TaurusBaseComponent.setModelInConfig`
        '''
        if yesno == self._modelInConfig:
            return
        if yesno:
            self.registerConfigProperty(
                self.getUseParentModel, self.setUseParentModel, "useParentModel")
        else:
            self.unregisterConfigurableItem(
                "useParentModel", raiseOnError=False)

        TaurusBaseComponent.setModelInConfig(self, yesno)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Mandatory overwrite from TaurusBaseComponent
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def updateStyle(self):
        """Updates the widget style. Default implementation just calls QWidget.update()

        Override when necessary.
        """
        self.update()

    def getParentTaurusComponent(self):
        """Returns the first taurus component in the widget hierarchy or None if no
        taurus component is found

        :return: (TaurusBaseComponent or None) the parent taurus base component
        """
        p = self.parentWidget()
        while p and not isinstance(p, TaurusBaseWidget):
            p = p.parentWidget()
        if isinstance(p, TaurusBaseWidget):
            self._parentTaurusComponent = p
        else:
            self._parentTaurusComponent = p = None
        return p

    def setDisconnectOnHide(self, disconnect):
        """Sets/unsets disconnection on hide event

        :param disconnect: (bool) whether or not to disconnect on hide event
        """
        if not self.visible() and disconnect == False:
            self.info(
                "Ignoring setDisconnectOnHide to False because widget is not visible")
            return
        self._disconnect_on_hide = disconnect

    def hideEvent(self, event):
        """Override of the QWidget.hideEvent()
        """
        if self._disconnect_on_hide:
            try:
                if self.getModelName():
                    self._detach()
                event.accept()
            except Exception:
                self.warning("Exception received while trying to hide")
                self.traceback()

    def showEvent(self, event):
        """Override of the QWidget.showEvent()"""
        if self._disconnect_on_hide:
            try:
                if self.getModelName():
                    self._attach()
                event.accept()
            except Exception:
                self.warning("Exception received while trying to show")
                self.traceback()

    def closeEvent(self, event):
        """Override of the QWidget.closeEvent()"""
        try:
            self._detach()
            event.accept()
        except Exception:
            self.warning("Exception received while trying to close")
            self.traceback()

    def handleException(self, e):
        for h in self._exception_listener:
            h.exceptionReceived(e)

    def _findSetTextMethod(self):
        """Determine if this widget is able to display the text value of the taurus
        model. It searches through the possible Qt methods to display text.

        :return: (callable) a python method or None if no suitable method is found.
        """
        setMethod = None
        try:
            setMethod = getattr(self, 'setText')
        except AttributeError:
            try:
                setMethod = getattr(self, 'setTitle')
            except AttributeError:
                try:
                    setMethod = getattr(self, 'display')
                except AttributeError:
                    # it seems the widget has no way to update a value
                    pass

        return setMethod

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Drag&Drop related methods:
    #    default implementation allows setting the model by dropping it on the
    #    widget (if the widget allows modifications by the user).
    #
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setModifiableByUser(self, modifiable):
        '''Reimplemented to acept/reject drops based on whether the widget is modifiable by the user.
        See :meth:`TaurusBaseComponent.setModifiableByUser()`'''
        TaurusBaseComponent.setModifiableByUser(self, modifiable)
        self.setAcceptDrops(modifiable)

    def getSupportedMimeTypes(self):
        '''
        returns a list of supported mimeTypes that this widget support (ordered
        by priority). If none is set explicitely via :meth:`setSupportedMimeTypes`,
        a best effort will be tried based on the model class

        ..seealso: :meth:`setSupportedMimeTypes`

        This provides only a very basic implementation. Reimplement in derived classes if needed

        :return: (list<str>) list of MIME type names
        '''
        if self._supportedMimeTypes is not None:
            return self._supportedMimeTypes
        # fallback guess based on modelclass
        try:
            modelclass = self.getModelClass()
        except:
            return []
        if modelclass == TaurusDevice:
            return [TAURUS_DEV_MIME_TYPE, TAURUS_MODEL_MIME_TYPE]
        elif modelclass == TaurusAttribute:
            return [TAURUS_ATTR_MIME_TYPE, TAURUS_MODEL_MIME_TYPE]
        else:
            return [TAURUS_MODEL_MIME_TYPE]

    def setSupportedMimeTypes(self, mimetypes):
        '''
        sets the mimeTypes that this widget support

        :param mimetypes: (list<str>) list (ordered by priority) of MIME type names
        '''
        self._supportedMimeTypes = mimetypes

    def dragEnterEvent(self, event):
        '''reimplemented to support drag&drop of models. See :class:`QWidget`'''
        if self.isModifiableByUser():
            supported = self.getSupportedMimeTypes()
            for f in event.mimeData().formats():
                if f in supported:
                    event.acceptProposedAction()
                    return

    def getDropEventCallback(self):
        '''returns the method to be called when a dropping event occurs.
        The default implementation returns `self.setModel`. Reimplement
        it subclasses to call different methods.

        :return: (callable)
        '''
        return self.setModel

    def dropEvent(self, event):
        '''reimplemented to support drag&drop of models. See :class:`QWidget`'''
        mtype = self.handleMimeData(
            event.mimeData(), self.getDropEventCallback())
        if mtype is None:
            self.info('Invalid model')
        else:
            event.acceptProposedAction()

    def handleMimeData(self, mimeData, method):
        '''Selects the most appropriate data from the given mimeData object
        (in the order returned by :meth:`getSupportedMimeTypes`) and passes
        it to the given method.

        :param mimeData: (QMimeData) the MIME data object from which the model
                         is to be extracted
        :param method: (callable<str>) a method that accepts a string as argument.
                       This method will be called with the data from the mimeData object

        :return: (str or None) returns the MimeType used if the model was
                 successfully set, or None if the model could not be set
        '''
        supported = self.getSupportedMimeTypes()
        formats = mimeData.formats()
        for mtype in supported:
            if mtype in formats:
                d = str(mimeData.data(mtype))
                if d is None:
                    return None
                try:
                    method(d)
                    return mtype
                except:
                    self.debug('Invalid data (%s) for MIMETYPE=%s' %
                               (repr(d), repr(mtype)))
                    self.traceback(taurus.Debug)
                    return None

    def getModelMimeData(self):
        '''Returns a MimeData object containing the model data. The default implementation
        fills the `TAURUS_MODEL_MIME_TYPE`. If the widget's Model class is
        Attribute or Device, it also fills `TAURUS_ATTR_MIME_TYPE` or
        `TAURUS_DEV_MIME_TYPE`, respectively

        :return: (QMimeData)
        '''
        mimeData = Qt.QMimeData()
        modelname = self.getModelName()
        mimeData.setData(TAURUS_MODEL_MIME_TYPE, modelname)
        try:
            modelclass = self.getModelClass()
        except:
            modelclass = None
        if issubclass(modelclass, TaurusDevice):
            mimeData.setData(TAURUS_DEV_MIME_TYPE, modelname)
        elif issubclass(modelclass, TaurusAttribute):
            mimeData.setData(TAURUS_ATTR_MIME_TYPE, modelname)
        return mimeData

    def mousePressEvent(self, event):
        '''reimplemented to record the start position for drag events.
        See :class:`~PyQt4.QtGui.QWidget`'''
        if self._dragEnabled and event.button() == Qt.Qt.LeftButton:
            # I need to copy it explicetely to avoid a bug with PyQt4.4
            self.dragStartPosition = Qt.QPoint(event.pos())
        self.getQtClass().mousePressEvent(self, event)

    def mouseMoveEvent(self, event):
        '''reimplemented to provide drag events.
        See :class:`~PyQt4.QtGui.QWidget`'''
        if not self._dragEnabled or not event.buttons() & Qt.Qt.LeftButton:
            return self.getQtClass().mouseMoveEvent(self, event)
        if (event.pos() - self.dragStartPosition).manhattanLength() < Qt.QApplication.startDragDistance():
            return self.getQtClass().mouseMoveEvent(self, event)
        ret = self.getQtClass().mouseMoveEvent(self, event)  # call the superclass
        event.accept()  # we make sure we accept after having called the superclass so that it is not propagated (many default implementations of mouseMoveEvent call event.ignore())
        drag = Qt.QDrag(self)
        drag.setMimeData(self.getModelMimeData())
        drag.exec_(Qt.Qt.CopyAction, Qt.Qt.CopyAction)
        return ret

    def isDragEnabled(self):
        '''whether the user can drag data from this widget

        :return: (bool) True if the user can drag data'''
        return self._dragEnabled

    def setDragEnabled(self, enabled):
        '''
        sets whether the user is allowed to drag data from this widget

        :param modifiable: (bool)
        '''
        self._dragEnabled = enabled

    def resetDragEnabled(self):
        '''Equivalent to setDragEnabled(self.__class__._dragEnabled)'''
        self.setModifiableByUser(self.__class__._dragEnabled)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Pending operations related methods: default implementation
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def updatePendingOpsStyle(self):
        """This method should be reimplemented by derived classes that want to
        change their appearance depending whether there are pending operations or not"""
        pass

    def emitValueChanged(self, *args):
        """Connect the specific XXXXChanged signals from derived classes to this
        method in order to have a unified signal which can be used by Taurus Widgets"""
        self.valueChangedSignal.emit()
        self.updatePendingOpsStyle()  # by default, update its own style

    def safeApplyOperations(self, ops=None):
        """Applies the given operations (or the pending operations if None passed)

        :param ops: (sequence<taurus.core.taurusoperation.TaurusOperation> or None) list of operations to apply.
                    If None is given (default) the component fetches the pending operations

        :return: (bool) False if the apply was aborted by the user or if the
                 widget is in design mode. True otherwise.
        """

        if ops is None:
            ops = self.getPendingOperations()

        # Check if we need to take care of dangerous operations
        if self.getForceDangerousOperations():
            dangerMsgs = []
        else:
            dangerMsgs = [op.getDangerMessage()
                          for op in ops if len(op.getDangerMessage()) > 0]
        # warn the user if need be
        if len(dangerMsgs) == 1:
            result = Qt.QMessageBox.warning(self, "Potentially dangerous action",
                                            "%s\nProceed?" % dangerMsgs[0],
                                            Qt.QMessageBox.Ok | Qt.QMessageBox.Cancel,
                                            Qt.QMessageBox.Ok)
            if result != Qt.QMessageBox.Ok:
                return False

        elif len(dangerMsgs) > 1:
            warningDlg = Qt.QMessageBox(Qt.QMessageBox.Warning, " %d potentially dangerous actions" % len(dangerMsgs),
                                        "You are about to apply %d actions that may be potentially dangerous. Proceed?" % len(
                                            dangerMsgs),
                                        Qt.QMessageBox.Ok | Qt.QMessageBox.Cancel,
                                        self)
            details = "\n".join(dangerMsgs)
            warningDlg.setDetailedText(details)
            result = warningDlg.exec_()
            if result != Qt.QMessageBox.Ok:
                return False

        if self._designMode:
            self.info('Refusing to apply operation while in design mode')
            return False

        self.applyPendingOperations(ops)
        return True

    def setAutoTooltip(self, yesno):
        """Determines if the widget should automatically generate a tooltip
        based on the current widget model.

        :param yesno: (bool) True to automatically generate tooltip or False otherwise
        """
        self._autoTooltip = yesno

    def getAutoTooltip(self):
        """Returns if the widget is automatically generating a tooltip based
        on the current widget model.

        :return: (bool)  True if automatically generating tooltip or False otherwise
        """
        return self._autoTooltip

    @classmethod
    def getQtDesignerPluginInfo(cls):
        """Returns pertinent information in order to be able to build a valid
        QtDesigner widget plugin.

        The dictionary returned by this method should contain *at least* the
        following keys and values:
        - 'module' : a string representing the full python module name (ex.: 'taurus.qt.qtgui.base')
        - 'icon' : a string representing valid resource icon (ex.: 'designer:combobox.png')
        - 'container' : a bool telling if this widget is a container widget or not.

        This default implementation returns the following dictionary::

            { 'group'     : 'Taurus [Unclassified]',
              'icon'      : 'logos:taurus.png',
              'container' : False }

        :return: (dict) a map with pertinent designer information"""
        return {
            'group': 'Taurus [Unclassified]',
            'icon': 'logos:taurus.png',
            'container': False}


class TaurusBaseWritableWidget(TaurusBaseWidget):
    """The base class for all taurus input widgets

    it emits the applied signal when the value has been applied.
    """

    applied = baseSignal('applied')

    def __init__(self, name, taurus_parent=None, designMode=False):
        self.call__init__(TaurusBaseWidget, name,
                          parent=taurus_parent, designMode=designMode)

        self._lastValue = None

        # Overwrite not to show quality by default
        self._showQuality = False

        # Don't do auto-apply by default
        self._autoApply = False

        # Don't force a writing to attribute when there are not pending
        # operations
        self._forcedApply = False

        self.valueChangedSignal.connect(self.updatePendingOperations)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getModelClass(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        return TaurusAttribute

    def isReadOnly(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        return False

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setAutoApply(self, auto):
        '''
        Sets autoApply mode. In autoApply mode, the widget writes the value
        automatically whenever it is changed by the user (e.g., when
        :meth:`notifyValueChanged` is called). If False, a value changed just
        flags a "pending operation" which needs to be applied manually by
        the user before the value gets written.

        :param auto: (bool) True for setting autoApply mode. False for disabling
        '''
        self._autoApply = auto

    def getAutoApply(self):
        '''whether autoApply mode is enabled or not.

        :return: (bool)
        '''
        return self._autoApply

    def resetAutoApply(self):
        '''resets the autoApply mode (i.e.: sets it to False)'''
        self.setAutoApply(False)

    def setForcedApply(self, forced):
        '''Sets the forcedApply mode. In forcedApply mode, values are written even
        if there are not pending operations (e.g. even if the displayed value is
        the same as the currently applied one).

        .. seealso: :meth:`forceApply` and :meth:`writeValue`

        :param forced: (bool) True for setting forcedApply mode. False for disabling
        '''
        self._forcedApply = forced

    def getForcedApply(self):
        '''whether forcedApply mode is enabled or not.

        :return: (bool)
        '''
        return self._forcedApply

    def resetForcedApply(self):
        '''resets the forcedApply mode (i.e.: sets it to False)'''
        self.setForcedApply(False)

    @deprecation_decorator(rel='4.0', alt='notifyValueChanged')
    def valueChanged(self, *args):
        return self.notifyValueChanged(*args)

    def notifyValueChanged(self, *args):
        '''Subclasses should connect some particular signal to this method for
        indicating that something has changed.
        e.g., a QLineEdit should connect its "textChanged" signal...
        '''
        self.emitValueChanged()
        if self._autoApply:
            self.writeValue()

    def writeValue(self, forceApply=False):
        '''Writes the value to the attribute, either by applying pending
        operations or (if the ForcedApply flag is True), it writes directly when
        no operations are pending

        It emits the applied signal if apply is not aborted.

        :param forceApply: (bool) If True, it behaves as in forceApply mode
                           (even if the forceApply mode is disabled by
                           :meth:`setForceApply`)
        '''
        if self.hasPendingOperations():
            applied = self.safeApplyOperations()
            if applied:
                self.applied.emit()
            return

        # maybe we want to force an apply even if there are no pending ops...
        kmods = Qt.QCoreApplication.instance().keyboardModifiers()
        controlpressed = bool(kmods & Qt.Qt.ControlModifier)
        if self.getForcedApply() or forceApply or controlpressed:
            self.forceApply()

    def forceApply(self):
        '''It (re)applies the value regardless of pending operations.
        WARNING: USE WITH CARE. In most cases what you need is to make sure
        that pending operations are properly created, not calling this method

        It emits the applied signal if apply is not aborted.

        .. seealso: :meth:`setForceApply` and :meth:`writeValue`

        '''
        try:
            v = self.getValue()
            op = WriteAttrOperation(self.getModelObj(), v,
                                    self.getOperationCallbacks())
            op.setDangerMessage(self.getDangerMessage())
            applied = self.safeApplyOperations([op])
            if applied:
                self.applied.emit()
            self.info('Force-Applied value = %s' % str(v))
        except:
            self.error('Unexpected exception in forceApply')
            self.traceback()

    def handleEvent(self, src, evt_type, evt_value):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        if evt_type in (TaurusEventType.Change, TaurusEventType.Periodic):
            self.emitValueChanged()

    def postAttach(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        TaurusBaseWidget.postAttach(self)
        if self.isAttached():
            try:
                v = self.getModelValueObj().wvalue
            except:
                v = None
            self.setValue(v)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Pending operations related methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def resetPendingOperations(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        if self.isAttached():
            try:
                v = self.getModelValueObj().wvalue
            except:
                v = None
            self.setValue(v)
        TaurusBaseWidget.resetPendingOperations(self)
        self.updateStyle()

    def updatePendingOperations(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        model = self.getModelObj()
        try:
            model_value = model.getValueObj().wvalue
            wigdet_value = self.getValue()
            if model.areStrValuesEqual(model_value, wigdet_value):
                self._operations = []
            else:
                operation = WriteAttrOperation(model, wigdet_value,
                                               self.getOperationCallbacks())
                operation.setDangerMessage(self.getDangerMessage())
                self._operations = [operation]
        except:
            self._operations = []
        self.updateStyle()

    def getOperationCallbacks(self):
        '''returns the operation callbacks (i.e., a sequence of methods that will be called after an operation is executed
           (this default implementation it returns an empty list).

        :return: (sequence<callable>)
        '''
        return []

    def getValue(self):
        '''
        This method must be implemented in derived classes to return
        the value to be written. Note that this may differ
        from the displayed value (e.g. for a numeric value being
        edited by a QLineEdit-based widget, the displayed value will
        be a string while getValue will return a number)
        '''
        raise NotImplementedError(
            "Not allowed to call TaurusBaseWritableWidget.getValue()")

    def setValue(self, v):
        '''
        This method must be implemented in derived classes to provide
        a (widget-specific) way of updating the displayed value based
        on a given attribute value

        :param v: The attribute value
        '''
        raise NotImplementedError(
            "Not allowed to call TaurusBaseWritableWidget.setValue()")

    def updateStyle(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        if self._autoTooltip:
            toolTip = self.getFormatedToolTip()
            if self.hasPendingOperations():
                v_str = str(self.getValue())
                model_v_str = getattr(
                    self.getModelValueObj(), 'wvalue', '-----')
                toolTip += '<hr/>Displayed value (%s) differs from applied value (%s)' % (
                    v_str, model_v_str)
            self.setToolTip(toolTip)

    def _updateValidator(self, evt_value):  # TODO: Remove this method
        # re-set the validator ranges if applicable
        if evt_value is None:
            return
        v = self.validator()
        if isinstance(v, Qt.QIntValidator):
            bottom = evt_value.min_value
            top = evt_value.max_value
            bottom = int(
                bottom) if bottom != TaurusConfiguration.no_min_value else -sys.maxint
            top = int(
                top) if top != TaurusConfiguration.no_max_value else sys.maxint
            v.setRange(bottom, top)
            self.debug("Validator range set to %i-%i" % (bottom, top))
        elif isinstance(v, Qt.QDoubleValidator):
            bottom = evt_value.min_value
            top = evt_value.max_value
            bottom = float(
                bottom) if bottom != TaurusConfiguration.no_min_value else -float("inf")
            top = float(
                top) if top != TaurusConfiguration.no_max_value else float("inf")
            v.setBottom(bottom)
            v.setTop(top)
            self.debug("Validator range set to %f-%f" % (bottom, top))

    @classmethod
    def getQtDesignerPluginInfo(cls):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['group'] = 'Taurus Input'
        return ret



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

""" """

__all__ = ["Table"]

__docformat__ = "restructuredtext"


class Table:

    DefTermWidth = 80

    PrettyOpts = {'col_sep': ' |', 'col_head_sep': '-', 'border': '='}

    def __init__(self, elem_list, elem_fmt=None, term_width=None,
                 row_head_str=None, row_head_fmt='%-*s', row_head_width=None,
                 col_head_str=None, col_head_fmt='%*s',  col_head_width=None,
                 col_sep=' ', row_sep=' ', col_head_sep=None, border=None):

        self.nr_col = len(elem_list)
        self.nr_row = len(elem_list[0])
        self.elem_list = elem_list

        if elem_fmt is None:
            elem_fmt = self.nr_row * ['%*s']
        if len(elem_fmt) == 1:
            elem_fmt *= self.nr_row
        self.elem_fmt = elem_fmt

        self.term_width = term_width or Table.DefTermWidth
        self.col_sep = col_sep
        self.row_sep = row_sep
        self.col_head_sep = col_head_sep
        self.border = border

        max_len_fn = lambda x: reduce(max, map(len, x))

        self.row_head_str = row_head_str
        self.row_head_fmt = row_head_fmt
        if row_head_str is not None and len(row_head_str) != self.nr_row:
            msg = 'RowHeadStr nr (%d) and RowNr (%d) mistmatch' % \
                  (len(row_head_str), self.nr_row)
            raise ValueError, msg
        if row_head_width is None:
            if row_head_str is not None:
                row_head_width = max_len_fn(row_head_str)
            else:
                row_head_width = 0
        self.row_head_width = row_head_width

        self.col_head_str = col_head_str
        self.col_head_fmt = col_head_fmt
        if col_head_str is not None and len(col_head_str) != self.nr_col:
            msg = 'ColHeadStr nr (%d) and ColNr (%d) mistmatch' % \
                  len(col_head_str), self.nr_col
            raise ValueError, msg
        if col_head_width is None:
            if col_head_str is not None:
                col_head_width = reduce(max, map(max_len_fn, col_head_str))
            else:
                col_head_width = 10
        self.col_head_width = col_head_width
        if col_head_str is not None:
            self.col_head_lines = len(col_head_str[0])
        else:
            self.col_head_lines = 0

    def updateElem(self, elem_list):
        new_col, new_row = len(elem_list), len(elem_list[0])
        if new_col != self.nr_col or new_row != self.nr_row:
            raise 'Invalid new elem list size %dx%d, was %dx%d' % \
                  (new_col, new_row, self.nr_col, self.nr_row)
        self.elem_list = elem_list

    def genOutput(self, term_width=None):
        if term_width is None:
            term_width = self.term_width

        rhw, chw = self.row_head_width, self.col_head_width
        chl = self.col_head_lines
        lcs = len(self.col_sep)
        width = term_width - chw   # At least one disp column!
        if rhw > 0:
            width -= rhw + lcs
        disp_cols = width / (chw + lcs) + 1
        tot_width = chw + (disp_cols - 1) * (chw + lcs)
        tot_rows = chl + self.nr_row
        if rhw > 0:
            tot_width += rhw + lcs

        output = []

        if self.row_head_str is not None:
            row_head = []
            fmt = self.row_head_fmt
            for head in [''] * chl + self.row_head_str:
                head = fmt % (rhw, head)
                row_head.append(head + self.col_sep)
        else:
            row_head = [''] * tot_rows

        for i in xrange(0, self.nr_col, disp_cols):
            if i > 0:
                nr_sep = tot_width / len(self.row_sep)
                output.append(self.row_sep * nr_sep)

            row_end = min(i + disp_cols, self.nr_col)
            line = list(row_head)
            for j in xrange(i, row_end):
                elem = self.elem_list[j]
                if chl:
                    col_head = self.col_head_str[j]
                    if j > i:
                        for k in xrange(tot_rows):
                            line[k] += self.col_sep
                    fmt = self.col_head_fmt
                    for k in xrange(chl):
                        line[k] += fmt % (chw, col_head[k])

                for k in xrange(self.nr_row):
                    fmt = self.elem_fmt[k]
                    line[chl + k] += fmt % (chw, elem[k])

            max_width = reduce(max, map(len, line))
            if self.border is not None:
                nr_border = max_width / len(self.border)
                output.append(self.border * nr_border)
            for l in line[:chl]:
                output.append(l)
            if self.col_head_sep is not None:
                nr_sep = max_width / len(self.col_head_sep)
                output.append(self.col_head_sep * nr_sep)
            for l in line[chl:]:
                output.append(l)
            if self.border is not None:
                output.append(self.border * nr_border)

        return output



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Test for taurus.core being PyTango Independent"""

__docformat__ = 'restructuredtext'

from taurus.external import unittest
# import functools
# from taurus.test import insertTest
# from modulemanager import ModuleManager
#
# # Decorator
# isTangoAvailable = functools.partial(insertTest, helper_name='isTangoAvailable',
#                             blockPytango=False)
#
#
# *TODO
#@isTangoAvailable(blockPytango=True)
#@isTangoAvailable() # Default False
#@isTangoAvailable(blockPytango=True)


class CoreTangoIndependentTestCase(unittest.TestCase):
    '''Test the Tango-independent core. As part of the SEP3 specification
       Taurus's core must be functional without PyTango.
       This test checks that you can import taurus without PyTango.
    '''

    def _importTaurus(self):
        ''' Helper method'''
        try:
            import taurus
        except ImportError:
            raise ImportError("Cannot import Taurus")

    def test_basicImport(self):
        '''Check if Taurus can be imported without PyTango
        '''
        # skip if PyTango is available
        try:
            import PyTango
            msg = 'Cannot test Tango-independence if PyTango can be imported'
            self.skipTest(msg)
        except ImportError:
            pass
        # check that taurus can be imported
        self.assertRaises(ImportError, self._importTaurus())

#     _modmanager = ModuleManager()
#     def _basicImportWithoutPyTango(self):
#         '''Basic test just try to import taurus (without PyTango)
#         '''
#         # TODO:
#         # _basicImportWithoutPyTango -> test_basicImportWithoutPyTango
#         # We have problems blocking modules so this test failed
#         self._modmanager.blockModule('PyTango')
#         self.assertRaises(ImportError, self._importTaurus())
#
#     def isTangoAvailable(self, blockPytango=False):
#         '''Test to check is Tango module is available for Taurus
#         '''
#         # TODO:
#         # We have problems blocking modules so this test failed because that
#         if blockPytango:
#             self._modmanager.blockModule('PyTango')
#
#         from taurus.core.taurusmanager import TaurusManager
#         plugins = TaurusManager().buildPlugins()
#
#         if blockPytango:
#             print '\t***', plugins
#             msg = 'Taurus is using Tango scheme, but you are blocking PyTango'
#             self.assertFalse('tango' in plugins, msg)
#         else:
#             msg = 'Taurus can not use Tango scheme, maybe you have not' +\
#                   ' installed PyTango'
#             self.assertTrue('tango' in plugins, msg)
#
#
#     def tearDown(self):
#         ''' Restore the original PyTango module'''
#         # *TODO
#         #self._modmanager.reloadModule('PyTango')
#         pass
#



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module defines the TangoDevice object"""

__all__ = ["TangoDevice"]

__docformat__ = "restructuredtext"

import time
from PyTango import (DeviceProxy, DevFailed, LockerInfo, DevState)

from taurus.core.taurusdevice import TaurusDevice
from taurus.core.taurusbasetypes import (TaurusDevState, TaurusLockInfo,
                                         LockStatus, TaurusEventType)
from taurus.core.util.log import taurus4_deprecation


class _TangoInfo(object):

    def __init__(self):
        self.dev_class = self.dev_type = 'TangoDevice'
        self.doc_url = 'http://www.esrf.fr/computing/cs/tango/tango_doc/ds_doc/'
        self.server_host = 'Unknown'
        self.server_id = 'Unknown'
        self.server_version = 1


class TangoDevice(TaurusDevice):
    """A Device object representing an abstraction of the PyTango.DeviceProxy
       object in the taurus.core.tango scheme"""

    # helper class property that stores a reference to the corresponding
    # factory
    _factory = None
    _scheme = 'tango'
    _description = "A Tango Device"

    def __init__(self, name, **kw):
        """Object initialization."""
        self.call__init__(TaurusDevice, name, **kw)
        self._deviceObj = self._createHWObject()
        self._lock_info = TaurusLockInfo()
        self._deviceStateObj = None
        # TODO reimplement using the new codification
        self._deviceState = TaurusDevState.Undefined

    # Export the DeviceProxy interface into this object.
    # This way we can call for example read_attribute on an object of this
    # class
    def __getattr__(self, name):
        if self._deviceObj is not None:
            return getattr(self._deviceObj, name)
        cls_name = self.__class__.__name__
        raise AttributeError("'%s' has no attribute '%s'" % (cls_name, name))

    # def __setattr__(self, name, value):
    #     if '_deviceObj' in self.__dict__ and self._deviceObj is not None:
    #         return setattr(self._deviceObj, name, value)
    #     super(TaurusDevice, self).__setattr__(name, value)

    def __contains__(self, key):
        """delegate the contains interface to the device proxy"""
        hw = self.getDeviceProxy()
        if hw is None:
            return False
        return hw.__contains__(key)

    def __getitem__(self, key):
        """read attribute value using key-indexing syntax (e.g. as in a dict)
        on the device"""
        attr = self.getAttribute(key)
        return attr.read()

    def __setitem__(self, key, value):
        """set attribute value using key-indexing syntax (e.g. as in a dict)
        on the device"""
        attr = self.getAttribute(key)
        return attr.write(value)

    def getAttribute(self, attrname):
        """Returns the attribute object given its name"""
        slashnb = attrname.count('/')
        if slashnb == 0:
            attrname = "%s/%s" % (self.getFullName(), attrname)
        elif attrname[0] == '/':
            attrname = "%s%s" % (self.getFullName(), attrname)
        return self.factory().getAttribute(attrname)

    @taurus4_deprecation(alt='.stateObj.read().rvalue [Tango] or ' +
                           '.state [agnostic]')
    def getState(self, cache=True):
        stateAttrValue = self.stateObj.read(cache=cache)
        if not stateAttrValue is None:
            state_rvalue = stateAttrValue.rvalue
            return DevState.values[state_rvalue.value]
        return None

    @taurus4_deprecation(alt='.stateObj [Tango] or ' +
                           '.factory.getAttribute(state_full_name) [agnostic]')
    def getStateObj(self):
        return self.stateObj

    @taurus4_deprecation(alt="state")
    def getSWState(self, cache=True):
        raise Exception('getSWState has been removed. Use state instead')
        # return self.getValueObj().rvalue

    @property
    def state(self, cache=True):
        """Reimplemented from :class:`TaurusDevice` to use Tango's state
        attribute for diagnosis of the current state. It supports a "cache"
        kwarg

        :param cache: (bool) If True (default), cache will be used when reading
                      the state attribute of this device

        :return: (TaurusDevState)
        """
        self._deviceState = TaurusDevState.NotReady
        try:
            taurus_tango_state = self.stateObj.read(cache).rvalue
        except:
            try:
                if self.getDeviceProxy().import_info().exported:
                    self._deviceState = TaurusDevState.Undefined
                    return self._deviceState  # Undefined
                else:
                    return self._deviceState  # NotReady
            except:
                return self._deviceState  # NotReady
        from taurus.core.tango.enums import DevState as TaurusTangoDevState
        if taurus_tango_state == TaurusTangoDevState.UNKNOWN:
            self._deviceState = TaurusDevState.Undefined
        elif taurus_tango_state not in (TaurusTangoDevState.FAULT,
                                      TaurusTangoDevState.DISABLE,
                                      TaurusTangoDevState.INIT):
            self._deviceState = TaurusDevState.Ready
        return self._deviceState

    @taurus4_deprecation(alt="state [agnostic] or stateObj.read [Tango]")
    def getValueObj(self, cache=True):
        """ Deprecated by TEP14.
        ..warning::
            this bck-compat implementation is not perfect because the
            rvalue of the returned TangoAttributeValue is now a member of
            TaurusDevState instead of TaurusSWDevState
        """
        from taurus.core.tango.tangoattribute import TangoAttrValue
        ret = TangoAttrValue()
        ret.rvalue = self.state(cache)
        return ret

    def getDisplayDescrObj(self, cache=True):
        desc_obj = super(TangoDevice, self).getDisplayDescrObj(cache)
        # extend the info on dev state
        ret = []
        for name, value in desc_obj:
            if name.lower() == 'device state' and self.stateObj is not None:
                tg_state = self.stateObj.read(cache).rvalue.name
                value = "%s (%s)" % (value, tg_state)
            ret.append((name, value))
        return ret

    def cleanUp(self):
        self.trace("[TangoDevice] cleanUp")
        self._descr = None

        if not self._deviceStateObj is None:
            self._deviceStateObj.removeListener(self)
        self._deviceStateObj = None
        self._deviceObj = None
        TaurusDevice.cleanUp(self)

    @taurus4_deprecation(alt='.state().name')
    def getDisplayValue(self, cache=True):
        return self.state(cache).name

    def _createHWObject(self):
        try:
            return DeviceProxy(self.getFullName())
        except DevFailed, e:
            self.warning('Could not create HW object: %s' % (e[0].desc))
            self.traceback()

    @taurus4_deprecation(alt="getDeviceProxy()")
    def getHWObj(self):
        return self.getDeviceProxy()

    def getDeviceProxy(self):
        if self._deviceObj is None:
            self._deviceObj = self._createHWObject()
        return self._deviceObj

    @taurus4_deprecation(alt='getDeviceProxy() is not None')
    def isValidDev(self):
        """see: :meth:`TaurusDevice.isValid`"""
        return self._deviceObj is not None

    def lock(self, force=False):
        li = self.getLockInfo()
        if force:
            if self.getLockInfo().status == TaurusLockInfo.Locked:
                self.unlock(force=True)
        return self.getDeviceProxy().lock()

    def unlock(self, force=False):
        return self.getDeviceProxy().unlock(force)

    def getLockInfo(self, cache=False):
        lock_info = self._lock_info
        if cache and lock_info.status != LockStatus.Unknown:
            return lock_info
        try:
            dev = self.getDeviceProxy()
            li = LockerInfo()
            locked = dev.get_locker(li)
            msg = "%s " % self.getSimpleName()
            if locked:
                lock_info.id = pid = li.li
                lock_info.language = li.ll
                lock_info.host = host = li.locker_host
                lock_info.klass = li.locker_class
                if dev.is_locked_by_me():
                    status = LockStatus.LockedMaster
                    msg += "is locked by you!"
                else:
                    status = LockStatus.Locked
                    msg += "is locked by PID %s on %s" % (pid, host)
            else:
                lock_info.id = None
                lock_info.language = None
                lock_info.host = host = None
                lock_info.klass = None
                status = LockStatus.Unlocked
                msg += "is not locked"
            lock_info.status = status
            lock_info.status_msg = msg
        except:
            self._lock_info = lock_info = TaurusLockInfo()
        return lock_info

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Protected implementation
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def removeListener(self, listener):
        ret = TaurusDevice.removeListener(self, listener)
        if not ret or self.hasListeners():
            return ret  # False, None or True
        return self.stateObj.removeListener(self)

    def addListener(self, listener):
        weWereListening = self.hasListeners()
        ret = TaurusDevice.addListener(self, listener)
        if not ret:
            return ret
        # We are only listening to State if someone is listening to us
        if weWereListening:
            # We were listening already, so we must fake an event to the new
            # subscribed listener with the current value
            evt_value = self.__decode(self.stateObj.read())
            listeners = hasattr(listener, '__iter__') and listener or [
                listener]
            self.fireEvent(TaurusEventType.Change, evt_value, listeners)
        else:
            # We were not listening to events, but now we have to
            self.stateObj.addListener(self)
        return ret

    def eventReceived(self, event_src, event_type, event_value):
        if event_type == TaurusEventType.Config:
            return
        value = self.__decode(event_value)
        new_state = value.rvalue
        if new_state != self._deviceState:
            msg = "Device State changed %s -> %s" % (self._deviceState.name,
                                                     new_state.name)
            self.debug(msg)
            self._deviceState = new_state
            self.fireEvent(TaurusEventType.Change, value)

    def __decode(self, event_value):
        """Decode events from the state attribute into TangoAttrValues whose
        rvalue is the Device state"""
        from taurus.core.tango.tangoattribute import TangoAttrValue
        if isinstance(event_value, TangoAttrValue):  # for change events (&co)
            new_state = TaurusDevState.Ready
        elif isinstance(event_value, DevFailed):  # for error events
            new_state = TaurusDevState.NotReady
        else:
            self.info("Unexpected event value: %r", event_value)
            new_state = TaurusDevState.Undefined
        from taurus.core.taurusbasetypes import TaurusModelValue
        value = TaurusModelValue()
        value.rvalue = new_state
        return value

    def __pollResult(self, attrs, ts, result, error=False):
        if error:
            for attr in attrs.values():
                attr.poll(single=False, value=None, error=result, time=ts)
            return

        for da in result:
            if da.has_failed:
                v, err = None, DevFailed(*da.get_err_stack())
            else:
                v, err = da, None
            attr = attrs[da.name]
            attr.poll(single=False, value=v, error=err, time=ts)

    def __pollAsynch(self, attrs):
        ts = time.time()
        try:
            req_id = self.read_attributes_asynch(attrs.keys())
        except DevFailed as e:
            return False, e, ts
        return True, req_id, ts

    def __pollReply(self, attrs, req_id, timeout=None):
        ok, req_id, ts = req_id
        if not ok:
            self.__pollResult(attrs, ts, req_id, error=True)
            return

        if timeout is None:
            timeout = 0
        timeout = int(timeout * 1000)
        result = self.read_attributes_reply(req_id, timeout)
        self.__pollResult(attrs, ts, result)

    def poll(self, attrs, asynch=False, req_id=None):
        '''optimized by reading of multiple attributes in one go'''
        if req_id is not None:
            return self.__pollReply(attrs, req_id)

        if asynch:
            return self.__pollAsynch(attrs)

        error = False
        ts = time.time()
        try:
            result = self.read_attributes(attrs.keys())
        except DevFailed as e:
            error = True
            result = e
        self.__pollResult(attrs, ts, result, error=error)

    def _repr_html_(self):
        try:
            info = self.getDeviceProxy().info()
        except:
            info = _TangoInfo()
        txt = """\
<table>
    <tr><td>Short name</td><td>{simple_name}</td></tr>
    <tr><td>Standard name</td><td>{normal_name}</td></tr>
    <tr><td>Full name</td><td>{full_name}</td></tr>
    <tr><td>Device class</td><td>{dev_class}</td></tr>
    <tr><td>Server</td><td>{server_id}</td></tr>
    <tr><td>Documentation</td><td><a target="_blank" href="{doc_url}">{doc_url}</a></td></tr>
</table>
""".format(simple_name=self.getSimpleName(), normal_name=self.getNormalName(),
           full_name=self.getFullName(), dev_class=info.dev_class,
           server_id=info.server_id, doc_url=info.doc_url)
        return txt

    @taurus4_deprecation(alt=".description")
    def getDescription(self, cache=True):
        return self.description

    @property
    def description(self):
        try:
            self._description = self.getDeviceProxy().description()
        except:
            pass
        return self._description

    @property
    def stateObj(self):
        if self._deviceStateObj is None:
            self._deviceStateObj = self.getAttribute("state")
        return self._deviceStateObj



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This package provides taurus Qt models

Pure PyQt view based widgets can be used to display the contents of the several
model classes provided here.

Displaying the device list in a :class:`PyQt4.QtGui.QTreeView`::

    view = Qt.QTreeView()
    db = taurus.Database()
    model = taurus.qt.qtcore.TaurusDbDeviceModel()
    model.setDataSource(db)
    view.setModel(model)

Same example but in a :class:`PyQt4.QtGui.QTableView`::

    view = Qt.QTableView()
    db = taurus.Database()
    model = taurus.qt.qtcore.TaurusDbPlainDeviceModel()
    model.setDataSource(db)
    view.setModel(model)

And now inside a :class:`PyQt4.QtGui.QComboBox`::

    view = Qt.QComboBox()
    db = taurus.Database()
    model = taurus.qt.qtcore.TaurusDbPlainDeviceModel()
    model.setDataSource(db)
    view.setModel(model)

"""

__docformat__ = 'restructuredtext'

from taurusmodel import *
from taurusdatabasemodel import *



import imp
import importlib
from sys import modules
from taurus.core.util.singleton import Singleton


class ModuleManager(Singleton):
    '''ModuleManager class is an helper to manager the python modules.
       This class has methods to import, reload, and block python modules.
    '''

    def __init__(self):
        self._modules = {}

    def _deleteModule(self, modname):
        ''' Remove the given module name from the exported python modules

        :param modname: Module name
        :type modname: str
        '''
        try:
            thismod = modules[modname]
        except KeyError:
            # This module is not imported
            raise ValueError(modname)
        these_symbols = dir(thismod)

        del modules[modname]
        for mod in modules.values():
            try:
                delattr(mod, modname)
            except AttributeError:
                pass

    def blockModule(self, modname):
        ''' Remplace the given module name by None. inhibits the module.

        :param modname: Module name
        :type modname: str
        '''
        if self._modules.get(modname) is None:
            self.importModule(modname)

        if modules[modname] is not None:
            _mod = {}
            _mod['mod'] = modules[modname]
            _mod['submod'] = []
            # Delete and block submodules
            for mod in modules.keys():
                if mod.find(modname) != -1:
                    _mod['submod'].append((mod, modules[mod]))
                    self._deleteModule(mod)
                    modules[mod] = None
            # Delete and block the module
            self._deleteModule(modname)
            modules[modname] = None
            self._modules[modname] = _mod

    def reloadModule(self, modname):
        '''Reload the given module name.

        :param modname: Module name
        :type modname: str
        '''
        if self._modules.get(modname) is None:
            msg = 'ModuleManager: Trying to reload a not imported module, %s'\
                  % (modname)
            print msg
            return
        # Reload the submodules
        for subname, submod in self._modules[modname]['submod']:
            modules[subname] = submod
        # Reload the module
        modules[modname] = self._modules[modname]['mod']
        imp.reload(self._modules[modname]['mod'])

    def importModule(self, modname):
        ''' Import the given module name.

        :param modname: Module name
        :type modname: str
        '''
        try:
            mod = __import__(modname)
            modules[modname] = mod
            _mod = {}
            _mod['mod'] = mod
            _mod['submod'] = []
            self._modules[modname] = _mod
        except ImportError:
            print 'Imposible to import the module %s' % (modname)

    def _getModuleDict(self, modname):
        ''' Return a dictionary with the given module name and its submodels
            if exists or None.

        :param modname: Module name
        :type modname: str

        :return dictionary
        '''
        return self._modules.get(modname)

    def getModule(self, modname):
        ''' Return a module of the given module name if exists or None.

        :param modname: Module name
        :type modname: str

        :return python module
        '''
        d = self._getModuleDict(modname)
        if d is None:
            return None
        return d['mod']

#---------------------Just 4 Test-----------------------------------------


def blockPyTango(modmanager):
    import taurus
    from taurus.core.taurusmanager import TaurusManager
    tm = TaurusManager()
    print '\n-*-*-*-*-*- Block PyTango'
    modmanager.blockModule('PyTango')
    taurus.check_dependencies()
    print '\n -- Taurus plugins'
    print tm.buildPlugins()


def reloadPyTango(modmanager):
    import taurus
    from taurus.core.taurusmanager import TaurusManager
    tm = TaurusManager()
    print '\n-*-*-*-*-*- Reload PyTango'
    modmanager.reloadModule('PyTango')
    taurus.check_dependencies()
    print '\n -- Taurus plugins'
    print tm.buildPlugins()


def test():
    modmanager = ModuleManager()
    blockPyTango(modmanager)
    reloadPyTango(modmanager)
    blockPyTango(modmanager)


if __name__ == '__main__':
    test()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
taurusgraphicview.py:
"""

__all__ = ['TaurusGraphicsView']

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt

from taurus.qt.qtgui.base import TaurusBaseWidget


class TaurusGraphicsView(Qt.QGraphicsView, TaurusBaseWidget):

    def __init__(self, parent=None, designMode=False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QGraphicsView, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)
        self.defineStyle()

    def defineStyle(self):
        self.updateStyle()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget over writing
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def isReadOnly(self):
        return True

    def updateStyle(self):
        self.update()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.graphic'
        ret['group'] = 'Taurus Display'
        ret['icon'] = "designer:graphicsview.png"
        return ret



#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module provides an icon catalog widget
"""

import os
import hashlib
from taurus.qt.qtgui.application import TaurusApplication
from taurus.qt.qtgui.input import GraphicalChoiceWidget
from taurus.qt.qtgui.icon import REGISTERED_PREFIXES, getCachedPixmap
from taurus.external.qt import Qt


class QIconCatalogPage(GraphicalChoiceWidget):
    """A widget that shows all icons available under a given searchPath preffix
    """

    def __init__(self, prefix, iconSize=24, columns=10):
        choices, pixmaps = self.__build_catalog(prefix, columns=columns)
        GraphicalChoiceWidget.__init__(self, choices=choices, pixmaps=pixmaps,
                                       iconSize=iconSize, defaultPixmap=None)

        spacer = Qt.QSpacerItem(0, 0, Qt.QSizePolicy.Expanding,
                                Qt.QSizePolicy.Expanding)

        self.gridLayout.addItem(spacer, self.gridLayout.rowCount(),
                                self.gridLayout.columnCount())

    def __build_catalog(self, prefix, columns=10):
        """explores paths registered under the given prefix and selects unique
        pixmaps (performs an md5 check to discard duplicated icon files)
        """
        pixmaps_hashed = {}
        hashes = {}

        for path in Qt.QDir.searchPaths(prefix):
            if not os.path.exists(path):
                print " %s not found. Skipping.!" % path
                continue

            for fname in os.listdir(path):
                abs_fname = os.path.join(path, fname)
                if not os.path.isfile(abs_fname):
                    continue
                md5 = self.__md5(abs_fname)
                choice = '%s:%s' % (prefix, fname)

                if md5 in hashes:
                    hashes[md5] += '\n%s\t(%s)' % (choice, abs_fname)
                else:
                    hashes[md5] = '%s\t(%s)' % (choice, abs_fname)
                    pixmap = getCachedPixmap(choice)
                    if not pixmap.isNull():
                        pixmaps_hashed[md5] = pixmap
        pixmaps = {}
        choices = []
        row = []
        for md5, choice in hashes.items():
            try:
                pixmaps[choice] = pixmaps_hashed[md5]
            except KeyError:
                continue
            row.append(choice)
            if len(row) > columns:
                choices.append(row)
                row = []
        if len(row) > 0:
            choices.append(row)

        return choices, pixmaps

    def __md5(self, fname):
        """
        Extracts md5 sum of a file

        :param fname: (str) path to file
        :return: md5 hash
        """
        hash_md5 = hashlib.md5()
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def onClick(self):
        """Reimplemented :class:`GraphicalChoiceWidget`
        """
        # From all alternatives, extract the one with the shortest name
        chosen = self.sender().text()
        alts = chosen.splitlines()
        alts = sorted(alts, key=lambda s: len(s.split()[0]))
        name, absname = alts[0].split()

        # Store chosen name and emit signal using name
        self._chosen = name
        self.choiceMade.emit(name)

        # show a message dialob with more info on the selected icon
        dlg = Qt.QMessageBox()
        dlg.setWindowTitle(name)
        text = 'You can access the selected icon as:\n%s\n\n' % name
        text += 'Or, by absolute name:\n%s\n\n' % absname
        if len(alts) > 1:
            text += 'Other alternative names:\n\n' + '\n\n'.join(alts[1:])
        dlg.setText(text)
        dlg.setIconPixmap(getCachedPixmap(name, size=128))
        dlg.exec_()

class QIconCatalog(Qt.QTabWidget):
    """
    A widget that shows a tab for each registered search path prefix.
    In each tab, all icons available for the corresponding prefix are displayed.
    Clicking on an icon provides info on how to use it from a taurus
    application.
    """

    iconSelected = Qt.pyqtSignal(str)

    def __init__(self, parent=None):
        Qt.QTabWidget.__init__(self)
        nprefix = len(REGISTERED_PREFIXES)
        progress = Qt.QProgressDialog('Building icon catalog...', None,
                                      0, nprefix, self)

        progress.setWindowModality(Qt.Qt.WindowModal)

        for i, prefix in enumerate(sorted(REGISTERED_PREFIXES)):
            progress.setValue(i)
            page = QIconCatalogPage(prefix)
            page.choiceMade.connect(self.iconSelected)

            self.addTab(page, prefix)
        progress.setValue(nprefix)


def main():
    """launcher of QIconCatalog"""
    import sys
    app = TaurusApplication(sys.argv)
    w = QIconCatalog()
    w.setWindowTitle('Taurus Icon Catalog')
    w.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



'''
This code exposes a limitation in the useParentModel architecture of Taurus and
explains how to avoid/work-around it.

Note that the conditions for hitting this limitation are quite special and that
in most situations you will not hit it. But unfortunately it may happen with
designer-generated files, and then the recheckTaurusParent() workaround is useful:
call recheckTaurusParent for all designer created widgets that use TaurusParentModel.
You can do it right after calling the setupUi method.
'''

from taurus.external.qt import Qt
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.display import TaurusLabel
import sys

app = Qt.QApplication(sys.argv)

# The problem arises in some situations when the Taurus parenting is not the same
# as the Qt parenting. For demonstration we use 3 widgets:
# p <-- m <--c (the arrows indicate the *Qt* parent)
# note that "m" not being a Taurus widget, implies that the *Taurus* parent of c is p
# also note that we are not giving the parent in the constructor, but we rely on
# doing it later on when adding to layout

p = TaurusWidget()  # Taurus parent
m = Qt.QWidget()  # midle widget (non Taurus)
c = TaurusLabel()  # Taurus child

# here we call setUseParentModel before the parent is known!

c.setUseParentModel(True)

# we prepare the layouts

m.setLayout(Qt.QVBoxLayout())
p.setLayout(Qt.QVBoxLayout())

# Now, by adding the widgets to the layout we are actually reparenting them.
# The order in which we reparent determines success/failure:
# if we do m-->p and then c-->m, it works, but if we do it in the opposite
# order, we trigger the error.
# (i.e., if we had called "p.layout().addWidget(m)" it would work work)
m.layout().addWidget(c)
p.layout().addWidget(m)

# the problem arises because the Taurus ancestry of c is only checked when:
# a) c.setUseParentModel() is called (and it effectively changes something) or
# b) when c is Qt-reparented
##
# In this example, c is reparented before m, so when it checks for a Taurus
# ancestor, none is found. Then m is reparented and c now would find its right
# Taurus ancestor (p)... but c never gets notified!

# Manually calling c.recheckTaurusParent() after the parenting has changed is
# another way to work around this issue (you can test by uncommenting
# the following line):

# c.recheckTaurusParent()

p.setModel('sys/tg_test/1/state')
print 'p model:', p.getModelName()
print 'c model:', c.getModelName()


p.show()
sys.exit(app.exec_())



import PyTango
import sys
import math


class TaurusTest(PyTango.Device_4Impl):

    def __init__(self, cl, name):
        PyTango.Device_4Impl.__init__(self, cl, name)
        TaurusTest.init_device(self)

    def delete_device(self):
        print "[Device delete_device method] for device", self.get_name()

    def init_device(self):
        print "In ", self.get_name(), "::init_device()"
        self.set_state(PyTango.DevState.ON)
        self.get_device_properties(self.get_device_class())
        self._position = 50.0
        self._velocity = 20.0
        self._acceleration = 4.0
        self._simulation_mode = False
        self._abscissas = [x / 50.0 for x in xrange(1024)]
        self._curve = [math.sin(x) for x in self._abscissas]

    def always_executed_hook(self):
        print "In ", self.get_name(), "::always_excuted_hook()"

    def read_attr_hardware(self, data):
        print "In ", self.get_name(), "::read_attr_hardware()"

    def read_Position(self, attr):
        attr.set_value(self._position)

    def write_Position(self, attr):
        self._position = attr.get_write_value()

    def read_Velocity(self, attr):
        attr.set_value(self._velocity)

    def is_Velocity_allowed(self, req_type):
        if req_type == PyTango.AttReqType.WRITE_REQ:
            return True
        return self._velocity < 5

    def write_Velocity(self, attr):
        self._velocity = attr.get_write_value()

    def read_Acceleration(self, attr):
        attr.set_value(self._acceleration)

    def write_Acceleration(self, attr):
        self._acceleration = attr.get_write_value()

    def read_SimulationMode(self, attr):
        attr.set_value(self._simulation_mode)

    def write_SimulationMode(self, attr):
        self._simulation_mode = attr.get_write_value()

    def read_Abscissas(self, attr):
        attr.set_value(self._abscissas)

    def read_Curve(self, attr):
        attr.set_value(self._curve)

    def write_Curve(self, attr):
        self._curve = attr.get_write_value()

    def create_device_cb(self, device_name):
        print "About to create device", device_name

    def CreateTaurusTestDevice(self, device_name):
        klass = self.get_device_class()
        klass.create_device(device_name, cb=self.create_device_cb)

    def DeleteTaurusTestDevice(self, device_name):
        klass = self.get_device_class()
        klass.delete_device(device_name)


class TaurusTestClass(PyTango.DeviceClass):

    #    Class Properties
    class_property_list = {
    }

    #    Device Properties
    device_property_list = {
    }

    #    Command definitions
    cmd_list = {
        'CreateTaurusTestDevice':
            [[PyTango.DevString, 'device name'],
             [PyTango.DevVoid, ""]],
        'DeleteTaurusTestDevice':
            [[PyTango.DevString, 'device name'],
             [PyTango.DevVoid, ""]],
    }

    #    Attribute definitions
    attr_list = {
        'Position':
            [[PyTango.DevDouble,
              PyTango.SCALAR,
              PyTango.READ_WRITE],
             {
                'label': "Gap",
                'unit': "mm",
                'format': "%8.3f",
                'max value': 1000,
                'min value': -1000,
                'max alarm': 900,
                'min alarm': -900,
                'max warning': 800,
                'min warning': -800,
            }],
        'Velocity':
            [[PyTango.DevDouble,
              PyTango.SCALAR,
              PyTango.READ_WRITE],
             {
                'label': "Speed",
                'unit': "nm/s",
                'format': "%6.2f",
                'max value': 100,
                'min value': 0,
                'max alarm': 95,
                'min alarm': 5,
                'max warning': 90,
                'min warning': 10,
            }],
        'Acceleration':
            [[PyTango.DevDouble,
              PyTango.SCALAR,
              PyTango.READ_WRITE],
             {
                'label': "Acceleration",
                'unit': "nm/s/s",
                'format': "%6.2f",
                'max value': 100,
                'min value': 0,
                'max alarm': 95,
                'min alarm': 2,
                'max warning': 90,
                'min warning': 1,
            }],
        'SimulationMode':
            [[PyTango.DevBoolean,
              PyTango.SCALAR,
              PyTango.READ_WRITE],
             {
                'label': "Simulation mode",
            }],
        'Abscissas':
            [[PyTango.DevDouble,
              PyTango.SPECTRUM,
              PyTango.READ, 1024],
             {
                'label': "X values for Curve 1",
            }],
        'Curve':
            [[PyTango.DevDouble,
              PyTango.SPECTRUM,
              PyTango.READ_WRITE, 1024],
             {
                'label': "Curve 1",
            }],

    }

    def __init__(self, name):
        PyTango.DeviceClass.__init__(self, name)
        self.set_type(name)
        print "In TaurusTestClass  constructor"


if __name__ == '__main__':
    try:
        py = PyTango.Util(sys.argv)
        py.add_TgClass(TaurusTestClass, TaurusTest, 'TaurusTest')

        U = PyTango.Util.instance()
        U.server_init()
        U.server_run()

    except PyTango.DevFailed, e:
        print '-------> Received a DevFailed exception:', e
    except Exception, e:
        print '-------> An unforeseen exception occured....', e



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
.. currentmodule:: taurus.core.evaluation

Evaluation extension for taurus core model.

The evaluation extension is a special extension that provides evaluation
objects. The official scheme name is 'eval'.

The main usage for this extension is to provide a way of performing mathematical
evaluations with data from other sources.

The Evaluation Factory (:class:`EvaluationFactory`) uses the following object
naming for referring to attributes (:class:`EvaluationAttribute`):

    `eval:[//<authority>][@<evaluator>/][<subst>;]<expr>`

or the following for referring to evaluation devices
(:class:`EvaluationDevice`):

    `eval:[//<authority>]@<evaluator>`

or the following for referring to an evaluation authority
(:class:`EvaluationAuthority`):

    `eval://<authority>`

where:

    - The `<authority>` segment is optional (except when referring to an
      EvaluationDatabase). At this point, only `//localhost` is supported.

    - The `@<evaluator>` is optional (except when referring to an
      EvaluationDevice). If not given, it defaults to `DefaultEvaluator`.

    - `<expr>` is a mathematical expression (using python syntax)
      that may have references to other taurus **attributes** by enclosing them
      between `{` and `}`. Expressions will be evaluated by the evaluator device
      to which the attribute is assigned.

    - The evaluator device inherits from :class:`SafeEvaluator` which by default
      includes a large subset of mathematical functions from the :mod:`numpy`
      module.

    - `<evaluator>` is a unique identification name for the evaluator device
      object. This allows to use different evaluator objects which may have
      different symbols available for evaluation. As an alternative approach, it
      is also possible to use custom-made evaluator devices as long as they
      inherit from :class:`EvaluationDevice`. To use a custom-made evaluator,
      you should construct the `<evaluator>` as `<modulename>.<classname>`
      where:

          - `<modulename>` is a python module name and

          - `<classname>` is the name of a class in `<modulename>` that derives
            from :class:`EvaluationDevice`

      See :file:`<taurus>/core/evaluation/dev_example.py` for an example of a
      custom Evaluator

    - The optional `<subst>` segment is used to provide substitution symbols.
      `<subst>` is a semicolon-separated string of `<key>=<value>` strings.


Some examples of valid evaluation models are:

    - An attribute that multiplies a tango attribute by 2:

        `eval:2*{tango:a/b/c/d}`

    - Same as above, but using substitutions:

        `eval:k=2;a={tango:a/b/c/d};k*a`

    - An attribute that adds two tango attributes together (assuming that tango
      is set as the default scheme)

        `eval:{a/b/c/d}+{f/g/h/i}`

    - An attribute that generates an array of random values:

        `eval:rand(256)`

    - An attribute that adds noise to a tango image attribute:

        `eval:img={tango:sys/tg_test/1/short_image_ro};img+10*rand(*img.shape)`

    - A default evaluator device named `foo`:

        `eval:@foo`

    - A custom evaluator device (implemented as class `MyEval` in the `mymod`
      module):

        `eval:@mymod.MyClass`


.. note:: Previous to SEP3, a RFC3986 non-compliant syntax was used for the
          evaluation scheme (e.g., allowing names such as
          ``tango://db=foo;dev=bar;a*b?k=2;a={tango:a/b/c/d}``).
          This syntax is now deprecated and should not be used. Taurus will
          issue warnings if detected.
"""

from evalfactory import EvaluationFactory
from evalattribute import EvaluationAttribute
from evalauthority import EvaluationAuthority
from evaldevice import EvaluationDevice



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

'''
Examples on using the evaluation scheme for exposing arbitrary non-tango quantities as taurus attributes
'''

__all__ = ['FreeSpaceDevice']

from taurus.core.evaluation import EvaluationDevice
import os
import platform
import ctypes


class FreeSpaceDevice(EvaluationDevice):
    '''A simple example of usage of the evaluation scheme for
    creating a device that allows to obtain available space in a dir.

    Important: note that only those members listed in `_symbols` will be available
    '''
    _symbols = ['getFreeSpace']

    def getFreeSpace(self, dir):
        """ return free space (in bytes).
        (Recipe adapted from `http://stackoverflow.com/questions/51658`)
        """
        if platform.system() == 'Windows':
            free_bytes = ctypes.c_ulonglong(0)
            ctypes.windll.kernel32.GetDiskFreeSpaceExW(
                ctypes.c_wchar_p(dir), None, None, ctypes.pointer(free_bytes))
            return free_bytes.value
        else:
            s = os.statvfs(dir)
            return s.f_bsize * s.f_bavail


#=========================================================================
# Just for testing
#=========================================================================

def test1():
    import taurus
    # calculates free space in Mb
    a = taurus.Attribute(
        'eval:@taurus.core.evaluation.dev_example.FreeSpaceDevice/getFreeSpace("/")/1024/1024')
    print "Free space: %iMb" % a.read().value


def test2():
    import sys
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus.qt.qtgui.display import TaurusLabel
    app = TaurusApplication()

    w = TaurusLabel()
    # calculates free space in Mb
    attrname = 'eval:@taurus.core.evaluation.dev_example.FreeSpaceDevice/getFreeSpace("/")'

    w.setModel(attrname)

    w.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    test2()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a set of basic taurus widgets based on QCheckBox"""

__all__ = ["TaurusAttrListComboBox", "TaurusValueComboBox"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt

from taurus.core import DataType, TaurusEventType
from taurus.core.taurusattribute import TaurusAttribute
from taurus.qt.qtgui.base import TaurusBaseWidget, TaurusBaseWritableWidget
from taurus.core.util import eventfilters


class TaurusValueComboBox(Qt.QComboBox, TaurusBaseWritableWidget):
    '''This widget shows a combobox that offers a limited choice of values that
    can be set on an attribute.'''

    def __init__(self, parent=None, designMode=False):
        self._previousModelName = None
        self._lastValueByUser = None

        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QComboBox, parent)
        self.call__init__(TaurusBaseWritableWidget,
                          name, designMode=designMode)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Helper methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def _setCurrentIndex(self, index):
        bs = self.blockSignals(True)
        try:
            self.setCurrentIndex(index)
        finally:
            self.blockSignals(bs)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def preAttach(self):
        '''reimplemented from :class:`TaurusBaseWritableWidget`'''
        TaurusBaseWritableWidget.preAttach(self)
        self.currentIndexChanged.connect(self.writeIndexValue)
        self.applied.connect(self.writeValue)

    def postDetach(self):
        '''reimplemented from :class:`TaurusBaseWritableWidget`'''
        TaurusBaseWritableWidget.postDetach(self)
        try:
            self.currentIndexChanged.disconnect(self.writeIndexValue)
            self.applied.disconnect(self.writeValue)
        except TypeError:
            # In new style-signal if a signal is disconnected without
            # previously was connected it, it raises a TypeError
            pass

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWritableWidget overwriting / Pending operations
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getValue(self):
        """
        Get the value that the widget is displaying now, not the value
        of the attribute.
        """
        model = self.getModelObj()
        if model is None:
            return None
        dtype = model.type
        new_value = self.itemData(self.currentIndex())
        if new_value is None:
            return None

        if dtype == DataType.Integer:
            func = int
        elif dtype == DataType.Float:
            func = float
        elif dtype == DataType.String:
            func = str
        elif dtype == DataType.Boolean:
            func = bool
        else:
            return None
        new_value = Qt.from_qvariant(new_value, func)
        return new_value

    def setValue(self, value):
        """
        Set the value for the widget to display, not the value of the
        attribute.
        """
        index = self.findData(Qt.QVariant(value))
        self._setCurrentIndex(index)

    def updateStyle(self):
        '''reimplemented from :class:`TaurusBaseWritableWidget`'''
        if self.hasPendingOperations():
            self.setStyleSheet('TaurusValueComboBox {color: blue; }')
        else:
            self.setStyleSheet('TaurusValueComboBox {}')
        super(TaurusValueComboBox, self).updateStyle()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # signals, gui events... things related to "write" in the end
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    @Qt.pyqtSlot(int, name='currentIndexChanged')
    def writeIndexValue(self, index):
        '''slot called to emit a valueChanged signal when the currentIndex is changed'''
        self.emitValueChanged()
        if self.getAutoApply():
            self.applied.emit()

    def keyPressEvent(self, event):
        '''reimplemented to emit an 'applied()' signal when Enter (or Return)
        key is pressed'''
        if event.key() in [Qt.Qt.Key_Return, Qt.Qt.Key_Enter]:
            self.applied.emit()
            event.accept()
        else:
            return Qt.QComboBox.keyPressEvent(self, event)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusValueComboBox own interface
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setValueNames(self, names):
        '''
        Sets the correspondence between the values to be applied and their
        associated text to show in the combobox.

        :param names: (sequence<tuple>) A sequence of (name,value) tuples,
                      where each attribute value gets a name for display
        '''
        bs = self.blockSignals(True)
        self.clear()
        self.blockSignals(bs)
        self.addValueNames(names)

    def addValueNames(self, names):
        '''
        Add new value-name associations to the combobox.

        ... seealso: :meth:`setValueNames`

        :param names: (sequence<tuple>) A sequence of (name,value) tuples,
                      where each attribute value gets a name for display
        '''
        bs = self.blockSignals(True)
        try:
            for k, v in names:
                self.addItem(k, Qt.QVariant(v))

            # Ok, now we should see if the current value matches any
            # of the newly added names. This is kinda a refresh:
            mv = self.getModelValueObj()
            if mv is not None:
                self.setValue(mv.wvalue)
        finally:
            self.blockSignals(bs)

        self.emitValueChanged()

    def getValueString(self, value, default='UNKNOWN(%s)'):
        """Returns the corresponding name in the combobox out of a value
        (or a default value if not found).

        :param value: value to look up
        :param default: (str) value in case it is not found. It accepts
                        a '%s' placeholder which will be substituted with
                        str(value). It defaults to 'UNKNOWN(%s)'.
        """
        item = self.findData(Qt.QVariant(value))
        if item < 0:
            if '%s' in default:
                return default % str(value)
            else:
                return default
        return str(self.itemText(item))

    def teachDisplayTranslationToWidget(self, widget, default='UNKNOWN(%s)'):
        """
        Makes a label object change the displayed text by the corresponding
        value of the combobox. This is implemented for the general case
        and may be not what you expect in some cases (as for example, it
        fires a fake periodic event which may be problematic if these
        are being filtered out).
        """
        # We reimplement label.displayValue so that instead of the normal
        # value it displays the string it has associated in the combobox.
        widget.displayValue = lambda v: self.getValueString(v, default)

        # Simulate a first event. Otherwise the displayValue will be
        # the default, not modified by us
        model = widget.getModelObj()
        if model:
            widget.fireEvent(model, TaurusEventType.Periodic,
                             model.getValueObj())

    def setQModel(self, *args, **kwargs):
        '''access to :meth:`QCombobox.setModel`

        .. seealso: :meth:`setModel`
        '''
        Qt.QComboBox.setModel(self, *args, **kwargs)

    def setModel(self, m):
        '''Reimplemented from :meth:`TaurusBaseWritableWidget.setModel` '''
        if isinstance(m, Qt.QAbstractItemModel):
            self.warning(
                "Deprecation warning: use setQModel() if you want to set a Qt Item Model. The setModel() method is reserved for Taurus models")
            return Qt.QComboBox.setModel(self, m)
        ret = TaurusBaseWritableWidget.setModel(self, m)
        self.emitValueChanged()
        return ret

    @classmethod
    def getQtDesignerPluginInfo(cls):
        '''reimplemented from :class:`TaurusBaseWritableWidget`'''
        ret = TaurusBaseWritableWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.input'
        ret['icon'] = "designer:combobox.png"
        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString",
                            TaurusBaseWidget.getModel,
                            TaurusBaseWidget.setModel,
                            TaurusBaseWidget.resetModel)

    useParentModel = Qt.pyqtProperty("bool",
                                     TaurusBaseWidget.getUseParentModel,
                                     TaurusBaseWidget.setUseParentModel,
                                     TaurusBaseWidget.resetUseParentModel)

    autoApply = Qt.pyqtProperty("bool",
                                TaurusBaseWritableWidget.getAutoApply,
                                TaurusBaseWritableWidget.setAutoApply,
                                TaurusBaseWritableWidget.resetAutoApply)

    forcedApply = Qt.pyqtProperty("bool", TaurusBaseWritableWidget.getForcedApply,
                                  TaurusBaseWritableWidget.setForcedApply,
                                  TaurusBaseWritableWidget.resetForcedApply)


class TaurusAttrListComboBox(Qt.QComboBox, TaurusBaseWidget):
    """Combobox whose items reflect the items read from a 1D attribute of dtype
    str
    """

    def __init__(self, parent=None, designMode=False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QComboBox, parent)
        self.call__init__(TaurusBaseWidget, name)
        self.insertEventFilter(eventfilters.IGNORE_CONFIG)
        self.setSizeAdjustPolicy(Qt.QComboBox.AdjustToContents)
        self.defineStyle()
        self._lastAttrList = None

    def defineStyle(self):
        """Defines the initial style for the widget """
        self.updateStyle()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget over writing
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getModelClass(self):
        """reimplemented from :class:`TaurusBaseWidget`"""
        return TaurusAttribute

    def handleEvent(self, evt_src, evt_type, evt_value):
        """reimplemented from :class:`TaurusBaseWidget`"""
        if evt_type == TaurusEventType.Error:
            attrList = []
        elif evt_src is None or evt_value is None:
            attrList = []
        else:
            attrList = list(evt_value.rvalue)
            attrList.sort()
        if attrList != self._lastAttrList:
            self._lastAttrList = attrList
            self.clear()
            self.addItems(attrList)
            self.updateStyle()

    def updateStyle(self):
        """reimplemented from :class:`TaurusBaseWidget`"""
        self.update()

    def setQModel(self, *args, **kwargs):
        """access to :meth:`QAbstractItemView.setModel`

        .. seealso: :meth:`setModel`
        """
        return Qt.QAbstractItemView.setModel(self, *args, **kwargs)

    def setModel(self, m):
        """reimplemented from :class:`TaurusBaseWidget`"""
        if isinstance(m, Qt.QAbstractItemModel):
            self.warning(("Deprecation warning: use setQModel() if you" +
                          " want to set a Qt Item Model. The setModel()" +
                          " method is reserved for Taurus models"))
            return Qt.QAbstractItemView.setQModel(self, m)
        return TaurusBaseWidget.setModel(self, m)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        """reimplemented from :class:`TaurusBaseWidget`"""
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['group'] = 'Taurus Input'
        ret['module'] = 'taurus.qt.qtgui.input'
        ret['icon'] = "designer:combobox.png"
        return ret

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString", TaurusBaseWidget.getModel,
                            TaurusBaseWidget.setModel,
                            TaurusBaseWidget.resetModel)

    useParentModel = Qt.pyqtProperty("bool",
                                     TaurusBaseWidget.getUseParentModel,
                                     TaurusBaseWidget.setUseParentModel,
                                     TaurusBaseWidget.resetUseParentModel)


#####################################################################
# Testing
#####################################################################
def _taurusAttrListTest():
    """tests taurusAttrList. Model: an attribute containing a list of strings"""
    from taurus.qt.qtgui.application import TaurusApplication
    a = TaurusApplication()
    # model = sys.argv[1]
    # model = "eval:['foo','bar']"
    model = "sys/tg_test/1/string_spectrum"
    w = TaurusAttrListComboBox()
    w.setModel(model)
    w.show()
    return a.exec_()


def _taurusValueComboboxTest():
    from taurus.qt.qtgui.application import TaurusApplication
    """tests TaurusValueCombobox """
    # model = sys.argv[1]
    model = 'sys/tg_test/1/short_scalar'
    names = [
        ('name0', 0),
        ('name1', 1),
        ('name2', 2),
        ('name3', 3)
    ]
    a = TaurusApplication()
    w = TaurusValueComboBox()
    w.setModel(model)
    w.addValueNames(names)
    #w.autoApply = True
    w.show()
    return a.exec_()

if __name__ == '__main__':
    import sys
    # main = _taurusValueComboboxTest #uncomment to test TaurusValueCombobox
    main = _taurusAttrListTest  # uncomment to testtaurusAttrList
    sys.exit(main())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
  This module provides the :class:`TaurusFactory` base class that any valid
  Factory in Taurus must inherit.

  The Factory objects are the basic block for building and interacting with a
  given scheme in Taurus. They provide Taurus model objects (TaurusAuthority,
  TaurusDevice or TaurusAttribute) for a given taurus model name.

  Taurus model naming is URI based (see <https://tools.ietf.org/html/rfc3986>)

  All the standard components of an URI (scheme, authority, path, query and
  fragment) may be part of a model name, and they are separated as follows:

  <scheme>:<authority><path>?<query>#<fragment>


  The following are some points to consider when using and/or implementing
  schemes based on this Abstract class:

  - It is strongly recommended that the scheme component is always present
  explicitly in the model name, although a default scheme can be defined in
  :mod:`taurus.tauruscustomsettings` so that model names which do not explicit
  the scheme can be auto-completed.

  - The authority component (if present on a given name) must always begin by
  a double slash ('//'). (see
  <https://tools.ietf.org/html/rfc3986#section-3.2>)

  - The path component, if present, must start by a single slash ('/') (see
  <https://tools.ietf.org/html/rfc3986#section-3.3>)

"""

__all__ = ["TaurusFactory"]

__docformat__ = "restructuredtext"

import atexit
from taurusbasetypes import TaurusElementType
from taurusauthority import TaurusAuthority
from taurusdevice import TaurusDevice
from taurusattribute import TaurusAttribute
from taurusconfiguration import TaurusConfiguration, TaurusConfigurationProxy


class TaurusFactory(object):
    """The base class for valid Factories in Taurus."""

    schemes = ()  # reimplement in derived classes to provide the supported sche
    caseSensitive = True  # reimplement if your scheme is case insensitive

    DefaultPollingPeriod = 3000

    def __init__(self):
        atexit.register(self.cleanUp)
        self._polling_period = self.DefaultPollingPeriod
        self.polling_timers = {}
        self._polling_enabled = True

        import taurusmanager
        manager = taurusmanager.TaurusManager()
        self._serialization_mode = manager.getSerializationMode()
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for cleanUp at exit
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def cleanUp(self):
        """Reimplement if you need to execute code on program execution exit.
        Default implementation does nothing.
        """
        pass

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # API for serialization
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setSerializationMode(self, mode):
        """Sets the serialization mode for the system.

        :param mode: (TaurusSerializationMode) the new serialization mode
        """
        self._serialization_mode = mode

    def getSerializationMode(self):
        """Gives the serialization operation mode.

        :return: (TaurusSerializationMode) the current serialization mode
        """
        return self._serialization_mode

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Methods that must be implemented by the specific Factory
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getAuthority(self, name=None):
        """getAuthority(string db_name) -> taurus.core.taurusauthority.TaurusAuthority

        Obtain the object corresponding to the given authority name or the
        default authority if db_name is None.
        If the corresponding authority object already exists, the existing
        instance is returned. Otherwise a new instance is stored and returned.

        :param db_name: [in] authority name string. It should be formed like:
                           <scheme>://<authority>. If <scheme> is ommited then
                           it will use the default scheme. if db_name is None,
                           the default authority is used

        :return: a taurus.core.taurusauthority.TaurusAuthority object
        :raises: :TaurusException: if the given name is invalid.
        """
        raise NotImplementedError("getAuthority cannot be called for abstract"
                                  " TaurusFactory")

    def getDevice(self, dev_name, **kw):
        """getDevice(string dev_name) -> taurus.core.taurusdevice.TaurusDevice

        Obtain the object corresponding to the given device name. If the
        corresponding device already exists, the existing instance is returned.
        Otherwise a new instance is stored and returned.

        :param dev_name: [in] the device name string. It should be formed like:
                            <scheme>://<authority>/<device name>. If <scheme>
                            is ommited then it will use the default scheme.
                            If authority is ommited then it will use the
                            default authority for the scheme.

        :return: a taurus.core.taurusdevice.TaurusDevice object
        :raises: :TaurusException: if the given name is invalid.
        """
        raise NotImplementedError("getDevice cannot be called for abstract"
                                  " TaurusFactory")

    def getAttribute(self, attr_name):
        """getAttribute(string attr_name) -> taurus.core.taurusattribute.TaurusAttribute

        Obtain the object corresponding to the given attribute name.
        If the corresponding attribute already exists, the existing instance
        is returned. Otherwise a new instance is stored and returned.

        :param attr_name: [in] string attribute name

        :return: a taurus.core.taurusattribute.TaurusAttribute object
        :raises: :TaurusException: if the given name is invalid.
        """
        raise NotImplementedError("getAttribute cannot be called for abstract"
                                  " TaurusFactory")

    def getAuthorityNameValidator(self):
        raise NotImplementedError("getAuthorityNameValidator cannot be called"
                                  " for abstract TaurusFactory")

    def getDeviceNameValidator(self):
        raise NotImplementedError("getDeviceNameValidator cannot be called"
                                  " for abstract TaurusFactory")

    def getAttributeNameValidator(self):
        raise NotImplementedError("getAttributeNameValidator cannot be called"
                                  " for abstract TaurusFactory")

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Factory extension API
    # Override the following methods if you need to provide special classes for
    # special object types
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def registerAttributeClass(self, attr_name, attr_klass):
        pass

    def unregisterAttributeClass(self, attr_name):
        pass

    def registerDeviceClass(self, dev_klass_name, dev_klass):
        pass

    def unregisterDeviceClass(self, dev_klass_name):
        pass

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Generic methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def supportsScheme(self, scheme):
        """Returns whether the given scheme is supported by this factory

        :param scheme: (str) the name of the schem to be checked

        :return: (bool) True if the scheme is supported (False otherwise)
        """
        return scheme in self.shemes

    def findObject(self, absolute_name):
        """ Must give an absolute name"""
        if not absolute_name:
            return None
        obj = None
        cls = self.findObjectClass(absolute_name)
        if cls:
            obj = self.getObject(cls, absolute_name)
        return obj

    def getObject(self, cls, name):
        t4_msg = 'The TaurusConfiguration classes are deprecated in tep14'
        if issubclass(cls, TaurusAuthority):
            return self.getAuthority(name)
        elif issubclass(cls, TaurusDevice):
            return self.getDevice(name)
        elif issubclass(cls, TaurusAttribute):
            return self.getAttribute(name)
        # For backward compatibility
        elif issubclass(cls, TaurusConfiguration):
            self.deprecated(dep='TaurusConfiguration', alt='TaurusAttribute',
                            rel='4.0', dbg_msg=t4_msg)
            return self.getAttribute(name)
        elif issubclass(cls, TaurusConfigurationProxy):
            self.deprecated(dep='TaurusConfigurationProxy',
                            alt='TaurusAttribute',
                            rel='4.0', dbg_msg=t4_msg)
            return self.getAttribute(name)
        else:
            return None

    def changeDefaultPollingPeriod(self, period):
        if period > 0:
            self._polling_period = period

    def getDefaultPollingPeriod(self):
        return self._polling_period

    def isPollingEnabled(self):
        """Tells if the local tango polling is enabled

           :return: (bool) whether or not the polling is enabled
        """
        return self._polling_enabled

    def disablePolling(self):
        """Disable the application tango polling"""
        if not self.isPollingEnabled():
            return
        self._polling_enabled = False
        for period, timer in self.polling_timers.iteritems():
            timer.stop()

    def enablePolling(self):
        """Enable the application tango polling"""
        if self.isPollingEnabled():
            return
        for period, timer in self.polling_timers.iteritems():
            timer.start()
        self._polling_enabled = True

    def addAttributeToPolling(self, attribute, period, unsubscribe_evts=False):
        """Activates the polling (client side) for the given attribute with the
           given period (seconds).

           :param attribute: (taurus.core.tango.TangoAttribute) attribute name.
           :param period: (float) polling period (in seconds)
           :param unsubscribe_evts: (bool) whether or not to unsubscribe from events
        """
        raise NotImplementedError("addAttributeToPolling cannot be called"
                                  " for abstract TaurusFactory")

    def removeAttributeFromPolling(self, attribute):
        """Deactivate the polling (client side) for the given attribute. If the
           polling of the attribute was not previously enabled, nothing happens.

           :param attribute: (str) attribute name.
        """
        raise NotImplementedError("removeAttributeFromPolling cannot be"
                                  " called for abstract TaurusFactory")

    def __str__(self):
        return '{0}()'.format(self.__class__.__name__)

    def __repr__(self):
        return '{0}(schemes={1})'.format(self.__class__.__name__, ", ".join(self.schemes))

    def getValidTypesForName(self, name, strict=None):
        '''
        Returns a list of all Taurus element types for which `name` is a valid
        model name (while in many cases a name may only be valid for one
        element type, this is not necessarily true in general)

        In this base implementation, name is checked first for Attribute, then
        for Device and finally for Authority, and the return value is sorted in
        that same order.

        If a given schema requires a different ordering, reimplement this method

        :param name: (str) taurus model name

        :return: (list<TaurusElementType.element>) where element can be one of:
                 `Attribute`, `Device` or `Authority`
        '''
        ret = []
        if self.getAttributeNameValidator().isValid(name, strict=strict):
            ret.append(TaurusElementType.Attribute)
        if self.getDeviceNameValidator().isValid(name, strict=strict):
            ret.append(TaurusElementType.Device)
        if self.getAuthorityNameValidator().isValid(name, strict=strict):
            ret.append(TaurusElementType.Authority)
        return ret

    def findObjectClass(self, absolute_name):
        """
        Obtain the class object corresponding to the given name.

        Note, this generic implementation expects that derived classes provide a
        an attribute called elementTypesMap consisting in a dictionary whose
        keys are TaurusElementTypes and whose values are the corresponding
        specific object classes. e.g., the FooFactory should provide::

          class FooFactory(TaurusFactory):
              elementTypesMap = {TaurusElementType.Authority: FooAuthority,
                                 TaurusElementType.Device: FooDevice,
                                 TaurusElementType.Attribute: FooAttribute,
                                 }
              (...)


        :param absolute_name: (str) the object absolute name string

        :return: (taurus.core.taurusmodel.TaurusModel or None) a TaurusModel
                 class derived type or None if the name is not valid

        """
        try:
            elementTypesMap = self.elementTypesMap
        except AttributeError:
            msg = ('generic findObjectClass called but %s does ' +
                   'not define elementTypesMap.') % self.__class__.__name__
            raise RuntimeError(msg)
        for t in self.getValidTypesForName(absolute_name):
            ret = elementTypesMap.get(t, None)
            if ret is not None:
                return ret
        return None
